"""
Package for the get_argument_to_typevar function.
"""
# Requires version-wrangling as typing changed between 3.6 and 3.7
from ...version import Python36Checker as __Python36Checker
if __Python36Checker().current_passes():
    from ._get_argument_to_typevar_3_6 import get_argument_to_typevar
else:
    from ._get_argument_to_typevar_3_7 import get_argument_to_typevar



def Tag(tag: str):
    """
    Decorator that adds the given tag to the decorated function.

    :param tag:     The tag to add to the function.
    """
    def apply_tag(method):
        if not hasattr(method, '__tags'):
            method.__tags = set()
        method.__tags.add(tag)
        return method

    return apply_tag


def has_tag(method, tag: str) -> bool:
    """
    Checks if the given method has the given tag.

    :param method:  The method to check.
    :param tag:     The tag to check for.
    :return:        True if the tag exists on the method,
                    False if not.
    """
    return hasattr(method, '__tags') and tag in method.__tags



from abc import abstractmethod, ABC
from weakref import WeakKeyDictionary

from .._OptionallyPresent import OptionallyPresent
from .._Absent import Absent
from ..._typing import RawJSONElement
from ...validator import JSONValidatorInstance


class Property(JSONValidatorInstance, ABC):
    """
    A property of a configuration is an instance value in the configuration
    object's keys. The value of a property is a raw JSON element, or something
    that can be converted to/from raw JSON. The property name can be set to the
    empty string to inherit the attribute name it is set against in the
    configuration.
    """
    def __init__(self,
                 name: str,
                 *,
                 optional: bool = False):
        self.__name: str = name
        self.__optional: bool = optional

        # The instance values of this property
        self.__values = WeakKeyDictionary()

    def name(self) -> str:
        """
        Gets the name of this property.

        :return:    The property name.
        """
        return self.__name

    def is_optional(self) -> bool:
        """
        Whether this property is an optional property.
        """
        return self.__optional

    def __get__(self, instance, owner):
        # If accessed from the class, return the property itself
        if instance is None:
            return self

        # Make sure a value has been set for this property
        if instance not in self.__values:
            raise AttributeError(f"{instance.__class__.__name__} has no value for property '{self.name()}'")

        # Get the value for this property
        return self.__values[instance]

    def __set__(self, instance, value):
        # Must be accessed via an instance
        self._require_instance(instance)

        # Validate the value
        self.validate_value(value)

        # Set the value in the instance dictionary
        self.__values[instance] = value

    @abstractmethod
    def get_as_raw_json(self, instance) -> OptionallyPresent[RawJSONElement]:
        """
        Gets the value of this property as raw JSON.

        :param instance:    The instance to get the value for.
        :return:            The raw JSON, or Absent.
        """
        pass

    @abstractmethod
    def set_from_raw_json(self, instance, value: OptionallyPresent[RawJSONElement]):
        """
        Sets the value of this property from raw JSON.

        :param instance:    The instance to set the value for.
        :param value:       The raw JSON value.
        """
        pass

    def __delete__(self, instance):
        raise AttributeError("Cannot delete configuration properties")

    def __set_name__(self, owner, name: str):
        # Can't have private properties
        if name.startswith("_"):
            raise AttributeError(f"Property names can't start with underscores ({name})")

        # Default the property name to the attribute name if one wasn't set
        if self.__name == "":
            self.__name = name

    @staticmethod
    def _require_instance(instance):
        """
        Raises an exception if instance is None.

        :param instance:    The instance.
        """
        # Make sure we are accessed via an instance
        if instance is None:
            raise AttributeError("Cannot access property through class (requires instance)")

    def validate_value(self, value):
        """
        Performs property value validation. Raises an AttributeError if validation fails.

        :param value:           The value to validate.
        """
        # Check for an absent value if allowed
        if not self.__optional and value is Absent:
            raise AttributeError(f"Cannot set non-optional property {self.name()} as absent")



import json
from abc import abstractmethod
from typing import TypeVar, Generic, IO

from ._error import JSONSerialiseError
from .._typing import RawJSONElement
from ..validator import JSONValidatorClass

# The type of the serialisable object
T = TypeVar("T", bound="JSONDeserialisable")


class JSONDeserialisable(Generic[T]):
    """
    Interface class for objects which can deserialise themselves from JSON.
    """
    @classmethod
    @abstractmethod
    def _deserialise_from_raw_json(cls, raw_json: RawJSONElement) -> T:
        """
        Implements the actual deserialisation from JSON.

        :param raw_json:    The raw JSON representation.
        :return:            An instance of the type.
        """
        pass

    @classmethod
    def from_raw_json(cls, raw_json: RawJSONElement) -> T:
        """
        Instantiates an object of this type from a raw JSON element.

        :param raw_json:    The raw JSON element.
        :return:            The object instance.
        """
        try:
            # Validate the raw JSON if we are capable
            if issubclass(cls, JSONValidatorClass):
                cls.validate_raw_json(raw_json)

            # Deserialise
            return cls._deserialise_from_raw_json(raw_json)
        except Exception as e:
            raise JSONSerialiseError("Error dumping JSON to string") from e

    @classmethod
    def from_json_string(cls, json_string: str) -> T:
        """
        Instantiates an object of this type from a JSON-format string.

        :param json_string:     The JSON-format string to parse.
        :return:                The object instance.
        """
        try:
            raw_json = json.loads(json_string)
        except Exception as e:
            raise JSONSerialiseError(f"Error parsing JSON string: {json_string}") from e

        return cls.from_raw_json(raw_json)

    @classmethod
    def read_from_stream(cls, stream: IO[str]) -> T:
        """
        Instantiates an object of this type from the given string-stream.

        :param stream:  The stream to read from.
        :return:        The object instance.
        """
        try:
            return cls.from_raw_json(json.load(stream))
        except Exception as e:
            raise JSONSerialiseError("Error reading state from stream") from e

    @classmethod
    def load_from_json_file(cls, filename: str) -> T:
        """
        Loads an instance of this class from the given file.

        :param filename:    The name of the file to load from.
        :return:            The instance.
        """
        try:
            with open(filename, 'r') as file:
                return cls.read_from_stream(file)
        except Exception as e:
            raise JSONSerialiseError(f"Could not load from JSON file '{filename}'") from e



import json
from abc import abstractmethod
from typing import IO

from ._error import JSONSerialiseError
from .._typing import RawJSONElement
from ..validator import JSONValidatorInstance


class JSONSerialisable:
    """
    Interface class for objects which can serialise themselves to JSON.
    """
    @abstractmethod
    def _serialise_to_raw_json(self) -> RawJSONElement:
        """
        Performs the actual serialisation.

        :return:    The JSON representation.
        """
        pass

    def to_raw_json(self) -> RawJSONElement:
        """
        Converts the state of this object to a raw JSON element.

        :return:    The raw JSON.
        """
        try:
            # Get the raw JSON representation
            raw_json: RawJSONElement = self._serialise_to_raw_json()

            # Validate the JSON if we are capable
            if isinstance(self, JSONValidatorInstance):
                self.validate_raw_json(raw_json)

            return raw_json
        except Exception as e:
            raise JSONSerialiseError("Error converting object to raw JSON") from e

    def to_json_string(self) -> str:
        """
        Serialises this object to a JSON-format string.

        :return:    The JSON string.
        """
        try:
            return json.dumps(self.to_raw_json())
        except Exception as e:
            raise JSONSerialiseError("Error converting object to JSON string") from e

    def write_to_stream(self, stream: IO[str]) -> None:
        """
        Writes this object as JSON to a string-stream.

        :param stream:  The stream to write to.
        """
        try:
            json.dump(self.to_raw_json(), stream)
        except Exception as e:
            raise JSONSerialiseError("Error writing state to stream") from e

    def save_to_json_file(self, filename: str) -> None:
        """
        Saves this object to the given file.

        :param filename:    The name of the file to save to.
        """
        try:
            with open(filename, 'w') as file:
                self.write_to_stream(file)
        except Exception as e:
            raise JSONSerialiseError(f"Could not save to JSON file '{filename}'") from e



from typing import Type

from .property import ProxyProperty
from ._Configuration import Configuration


class SubConfiguration(ProxyProperty):
    """
    Property which validates a sub-configuration (essentially an object).
    """
    def __init__(self,
                 name: str,
                 configuration_type: Type[Configuration],
                 *,
                 optional: bool = False):
        super().__init__(
            name,
            configuration_type,
            optional=optional
        )



from .validator import JSONValidatorInstance
from .schema import JSONSchema


class BasicSchemaValidator(JSONValidatorInstance):
    """
    Basic wrapper for a JSON schema which presents it as a validator.
    """
    def __init__(self, schema: JSONSchema):
        self.__schema = schema

    def get_json_validation_schema(self) -> JSONSchema:
        return self.__schema



from abc import ABC
from typing import Type

from ..._typing import RawJSONElement
from .._OptionallyPresent import OptionallyPresent
from .._Absent import Absent
from ...schema import JSONSchema
from ...serialise import JSONValidatedBiserialisable
from ._Property import Property


class ProxyProperty(Property, ABC):
    """
    Property which takes values of a type that is serialisable
    to/from JSON.
    """
    def __init__(self,
                 name: str,
                 type: Type[JSONValidatedBiserialisable],
                 *,
                 optional: bool = False):
        super().__init__(
            name,
            optional=optional
        )

        self.__type: Type[JSONValidatedBiserialisable] = type

    def type(self) -> Type[JSONValidatedBiserialisable]:
        """
        Gets the type of this proxy-property.
        """
        return self.__type

    def get_as_raw_json(self, instance) -> OptionallyPresent[RawJSONElement]:
        # Get the proxy value
        value = self.__get__(instance, None)

        # If not absent, serialise it
        return value.to_raw_json() if isinstance(value, JSONValidatedBiserialisable) else value

    def set_from_raw_json(self, instance, value: OptionallyPresent[RawJSONElement]):
        # Deserialise the value
        self.__set__(instance, self.__type.from_raw_json(value) if value is not Absent else Absent)

    def get_json_validation_schema(self) -> JSONSchema:
        return self.__type.get_json_validation_schema()

    def validate_value(self, value):
        super().validate_value(value)

        # No need to continue validation if value is absent
        if value is Absent:
            return

        # Must be an instance of the correct type
        if not isinstance(value, self.__type):
            raise AttributeError(f"{value} is not an instance of {self.__type.__name__}")



from .._Absent import Absent
from ...schema import JSONSchema
from ..._typing import RawJSONElement
from .._OptionallyPresent import OptionallyPresent
from ._Property import Property


class RawProperty(Property):
    """
    Property which takes raw JSON values. Uses a schema for validation.
    """
    def __init__(self,
                 name: str,
                 schema: JSONSchema,
                 *,
                 optional: bool = False):
        super().__init__(name, optional=optional)

        self.__schema = schema

    def get_as_raw_json(self, instance) -> OptionallyPresent[RawJSONElement]:
        return self.__get__(instance, None)

    def set_from_raw_json(self, instance, value: OptionallyPresent[RawJSONElement]):
        return self.__set__(instance, value)

    def get_json_validation_schema(self) -> JSONSchema:
        return self.__schema

    def validate_value(self, value: RawJSONElement):
        super().validate_value(value)

        # No need to continue validation if value is absent
        if value is Absent:
            return

        # Validate the raw JSON with our schema
        try:
            self.validate_raw_json(value)
        except Exception as e:
            raise AttributeError(f"{value} failed schema validation") from e



from typing import Optional


class AbsentType:
    """
    Singleton class representing a property that is absent from
    a configuration.
    """
    # The singleton instance
    __instance: Optional['AbsentType'] = None

    # Create the instance the first time, then just return it
    def __new__(cls):
        if cls.__instance is None:
            cls.__instance = super().__new__(cls)

        return cls.__instance

    def __str__(self) -> str:
        return "Absent"


# Create the singleton instance for export
Absent = AbsentType()



from typing import Optional

from ...schema import string_schema
from ._RawProperty import RawProperty


class StringProperty(RawProperty):
    """
    Configuration property which validates a string.
    """
    def __init__(self,
                 name: str,
                 min_length: Optional[int] = None,
                 max_length: Optional[int] = None,
                 pattern: Optional[str] = None,
                 format: Optional[str] = None,
                 *,
                 optional: bool = False):
        super().__init__(
            name,
            string_schema(
                min_length,
                max_length,
                pattern,
                format
            ),
            optional=optional
        )



"""
Module for working with meta-data on objects in an iterable.
"""
from typing import Any, Iterable, Iterator

from .._typing import T
from ._ConstantIterator import ConstantIterator


def with_metadata(iterable: Iterable[T], key: str, value: Any) -> Iterator[T]:
    """
    Adds meta-data to each object in an iterator.

    :param iterable:    The object iterable.
    :param key:         The meta-data key.
    :param value:       The meta-data value.
    :return:            An iterator over the objects with added meta-data.
    """
    return zip_metadata(iterable, ConstantIterator[str](key), ConstantIterator[str](value))


def zip_metadata(iterable: Iterable[T], keys: Iterable[str], values: Iterable[Any]) -> Iterator[T]:
    """
    Adds meta-data to each object in an iterator.

    :param iterable:    The object iterable.
    :param keys:        The meta-data key iterable.
    :param values:      The meta-data iterable.
    :return:            An iterator over the objects with added meta-data.
    """
    from ..meta import with_metadata
    return (with_metadata(obj, key, value) for obj, key, value in zip(iterable, keys, values))


def get_metadata(iterable: Iterable, key: str) -> Iterator:
    """
    Iterates over the meta-data attached to the objects in
    the given iterable under the given key.

    :param iterable:    The object iterable.
    :param key:         The meta-data key.
    :return:            An iterator over the meta-data.
    """
    return unzip_metadata(iterable, ConstantIterator[str](key))


def unzip_metadata(iterable: Iterable, keys: Iterable[str]) -> Iterator:
    """
    Iterates over the meta-data attached to the objects in
    the given iterable under the given keys.

    :param iterable:    The object iterable.
    :param keys:        The meta-data key iterable.
    :return:            An iterator over the meta-data.
    """
    from ..meta import get_metadata
    return (get_metadata(obj, key) for obj, key in zip(iterable, keys))



from ._typing import RawJSONElement


def deep_copy(json: RawJSONElement) -> RawJSONElement:
    """
    Creates a deep-copy of the given raw JSON element.

    :param json:    The raw JSON element to copy.
    :return:        The copy.
    """
    if isinstance(json, dict):
        return {key: deep_copy(value) for key, value in json.items()}
    elif isinstance(json, list):
        return [deep_copy(array_element) for array_element in json]
    else:
        return json



"""
Provides memoisation of function calls.

For long-running functions, it is sometimes nice to store the result,
and then simply load it the next time the function is called with the same
parameters. This module uses pickle to store the result to disk, and to
load it again.

Instead of:

    result = func(*args, **kwargs)

you would call:

    factory = MemoiserFactory("/path/to/memo/files/")
    memo_func = factory(func)
    result = memo_func(*args, **kwargs)

The first call will run the function as usual, but subsequent calls will load
the result from disk instead.

N.B.:
    - All arguments to the function and the function's result must be picklable.
    - Function state/closures are not pickled. Should only be used with pure functions.
"""
import hashlib
import os
import pickle
from typing import Callable, Tuple, Dict, Any

# The hashing function to use
HASH = hashlib.sha256

# The extension for memo files
MEMO_EXTENSION = ".memo"


class MemoFile:
    """
    Class representing .memo files on disk as produced/consumed
    by the Memoiser class.
    """
    def __init__(self, directory: str, func: Callable, *args, **kwargs):
        self.pickled_func_and_args = pickle_function_and_arguments(func, *args, **kwargs)
        self.code_hash: bytes = hash_function_code(func)
        self.filename: str = normalise_directory(directory) + get_memo_filename(self.pickled_func_and_args)

    def exists(self) -> bool:
        """
        Whether this memo file exists on disk.

        :return:    True if the file exists, False if not.
        """
        return os.path.exists(self.filename)

    def save(self, result):
        """
        Saves the given result of the function to the file.

        :param result:      The result to save.
        """
        with open(self.filename, "wb") as file:
            file.write(self.pickled_func_and_args)
            file.write(self.code_hash)
            pickle.dump(result, file)

    def load(self):
        """
        Loads the result from this memo file. Raises an error if the
        result can't be loaded for any reason.

        :return:    The saved result.
        """
        try:
            with open(self.filename, "rb") as file:
                # Check that the file is valid
                self._check_stored_func_and_args(file)
                self._check_stored_code_hash(file)

                # Return the stored result
                return pickle.load(file)

        # Just re-raise any FNF or already-wrapped errors
        except (FileNotFoundError, InvalidMemoFileError):
            raise

        # Wrap any other errors
        except Exception as e:
            raise InvalidMemoFileError("Error reading " + self.filename + ": " + str(e)) from e

    def _check_stored_func_and_args(self, file):
        """
        Checks if the function/arguments stored in the given file
        are valid for this memo-file. Raises an error if not, and
        simply returns if they do.

        :param file:    The file handle.
        """
        try:
            stored_func_and_args = file.read(len(self.pickled_func_and_args))
        except Exception as e:
            raise InvalidMemoFileError("Error reading stored function/arguments from " +
                                       self.filename + ": " + str(e)) from e

        if self.pickled_func_and_args != stored_func_and_args:
            raise InvalidMemoFileError("Hash collision for " + self.filename)

    def _check_stored_code_hash(self, file):
        """
        Checks if the code hash stored in the given file
        is valid for this memo-file. Raises an error if not, and
        simply returns if it is.

        :param file:    The file handle.
        """
        try:
            stored_code_hash = file.read(len(self.code_hash))
        except Exception as e:
            raise InvalidMemoFileError("Error reading stored code hash from " +
                                       self.filename + ": " + str(e)) from e

        if self.code_hash == stored_code_hash:
            raise InvalidMemoFileError("Function code change invalidates " + self.filename)


class Memoiser:
    """
    Class that acts like the given function, except memoises the result
    to disk. If the function is then called with the same arguments,
    reads the result from disk instead.
    """
    def __init__(self, func: Callable, save_directory: str = ""):
        self.func: Callable = func
        self.save_directory: str = normalise_directory(save_directory)

        # Create the save directory if it doesn't exist yet
        if not os.path.exists(self.save_directory):
            os.mkdir(self.save_directory)

    def __call__(self, *args, **kwargs):
        # Get the memo file to use for this function/arguments
        memo_file: MemoFile = MemoFile(self.save_directory, self.func, *args, **kwargs)

        # Try to load the result
        try:
            if memo_file.exists():
                return memo_file.load()
        except InvalidMemoFileError:
            # Continue to normal function call if file loading fails
            pass

        # If loading failed, call the function
        result = self.func(*args, **kwargs)

        # Save the result for next time
        memo_file.save(result)

        # Return the result
        return result


class MemoiserFactory:
    """
    Class that produces Memoisers all using the same save directory.
    """
    def __init__(self, save_directory: str = ""):
        self.save_directory: str = normalise_directory(save_directory)

    def __call__(self, func: Callable) -> Memoiser:
        return Memoiser(func, self.save_directory)

    def sub_dir(self, path: str) -> 'MemoiserFactory':
        """
        Returns a factory for a sub-directory of this one.

        :param path:    The extended path of the sub-directory.
        :return:        The factory.
        """
        # Normalise the path
        path = normalise_directory(path)

        # Remove any leading separator
        if path.startswith(os.sep):
            path = path[1:]

        return MemoiserFactory(self.save_directory + path)


class InvalidMemoFileError(Exception):
    """
    Error for when a memo file is invalidated by a change in function code,
    or a hash-collision.
    """
    pass


def normalise_directory(directory: str) -> str:
    """
    Normalises a directory string so that a filename can
    be directly appended and used.

    :param directory:   The directory string to normalise.
    :return:            The normalised directory string.
    """
    # Add the final separator if not cwd and not already present
    if directory != "" and not directory.endswith(os.sep):
        directory += os.sep

    return directory


def get_memo_filename(data: bytes) -> str:
    """
    Generates a memo filename for a bytes.

    :param data:    The bytes.
    :return:        The filename.
    """
    # Hash the data and get the hex representation as a string
    data_hash = HASH(data).hexdigest()

    # Return the last 16 hex characters of the hash with the extension
    return data_hash[-16:] + MEMO_EXTENSION


def pickle_function_and_arguments(func: Callable, *args, **kwargs):
    """
    Creates a binary representation of a function and its arguments.

    :param func:    The function.
    :param args:    The positional arguments to the function.
    :param kwargs:  The keyword arguments to the function.
    :return:        The binary pickle representation.
    """
    return pickle.dumps((func, args, PicklableDict(kwargs)))


def hash_function_code(func: Callable) -> bytes:
    """
    Hashes the code of a function so that it can be identified
    when it changes.

    :param func:    The function to hash.
    :return:        The hash bytes.
    """
    return HASH(func.__code__.co_code).digest()


class PicklableDict:
    """
    Class which is a picklable representation of a dictionary
    (assuming all keys/values are picklable).
    """
    def __init__(self, dictionary: Dict[Any, Any]):
        # The internal representation of the dictionary
        self.intern: Tuple[Tuple[Any, Any], ...] = tuple(dictionary.items())

    def __eq__(self, other):
        if isinstance(other, PicklableDict):
            other = other.to_dict()

        if isinstance(other, dict):
            return other == self.to_dict()

        return False

    def to_dict(self) -> Dict[Any, Any]:
        """
        Reconstructs the dictionary.

        :return:    The dictionary.
        """
        return {key: value for key, value in self.intern}



"""
Package for working with JSON-format configuration files.
"""
from ._Absent import Absent
from ._Configuration import Configuration
from ._OptionallyPresent import OptionallyPresent
from ._StrictConfiguration import StrictConfiguration
from ._SubConfiguration import SubConfiguration



"""
Package for working with Python versions.
"""
from ._Python36Checker import Python36Checker
from ._Python37Checker import Python37Checker
from ._PythonVersionChecker import PythonVersionChecker, PythonVersionError
from ._SupportedVersionChecker import SupportedVersionChecker



"""
Package for reading and representing CSV files.

Typical usage:
from wai.common.file.csv import loadf
file: CSVFile = loadf('filename.csv')
process_csv_file(file)
"""
from ._CSVFile import CSVFile
from ._load import load, loadf, loads
from ._save import save



from ...schema import BOOL_SCHEMA
from ._RawProperty import RawProperty


class BoolProperty(RawProperty):
    """
    Configuration property which validates a boolean value.
    """
    def __init__(self,
                 name: str,
                 *,
                 optional: bool = False):
        super().__init__(
            name,
            BOOL_SCHEMA,
            optional=optional
        )



from typing import Optional

from ...schema import number
from ..._typing import RawJSONNumber
from ._RawProperty import RawProperty


class NumberProperty(RawProperty):
    """
    Property which validates a number.
    """
    def __init__(self,
                 name: str,
                 minimum: Optional[RawJSONNumber] = None,
                 maximum: Optional[RawJSONNumber] = None,
                 integer_only: bool = False,
                 multiple_of: Optional[RawJSONNumber] = None,
                 exclusive_minimum: bool = False,
                 exclusive_maximum: bool = False,
                 *,
                 optional: bool = False):
        super().__init__(
            name,
            number(
                minimum,
                maximum,
                integer_only,
                multiple_of,
                exclusive_minimum,
                exclusive_maximum
            ),
            optional=optional
        )



from ._ARFFFile import ARFFFile, VALUE_TYPE, ROW_TYPE, DATA_TYPE, SELECTION_TYPE
from ._Attribute import Attribute, NumericAttribute, NominalAttribute, StringAttribute, DateAttribute, \
    InvalidNominalValueError, AttributeNameNotFoundError, AttributeTypeNotFoundError, InvalidNumericSubTypeError, \
    InvalidValueTypeError, NominalValuesError
from ._error import ARFFError
from ._load import load, loads, loadf, DataSizeMismatchError, RelationNameNotFoundError
from ._load_helper import UnrecognisedContentError, KeywordNotFoundError
from ._save import save



from typing import Optional

from ._Property import Property
from ._ProxyProperty import ProxyProperty
from ._ArrayProxy import ArrayProxy


class ArrayProperty(ProxyProperty):
    """
    Property which validates a regular array of elements.
    """
    def __init__(self,
                 sub_property: Property,
                 min_elements: int = 0,
                 max_elements: Optional[int] = None,
                 unique_elements: bool = False):

        # Create a closure class to proxy the array
        class ClosureArrayProxy(ArrayProxy):
            @classmethod
            def sub_property(cls) -> Property:
                return sub_property

            @classmethod
            def min_elements(cls) -> int:
                return min_elements

            @classmethod
            def max_elements(cls) -> Optional[int]:
                return max_elements

            @classmethod
            def unique_elements(cls) -> bool:
                return unique_elements

        super().__init__(
            sub_property.name(),
            ClosureArrayProxy,
            optional=sub_property.is_optional()
        )

    def __set__(self, instance, value):
        # Convert other proxy-arrays and raw lists to our proxy-type
        if (isinstance(value, ArrayProxy) and not isinstance(value, self.type())) or \
           isinstance(value, list):
            value = self.type()(value)

        super().__set__(instance, value)



from typing import List, Optional

from ._Report import Report, NO_ID
from ._SpectrumPoint import SpectrumPoint


class Spectrum:
    """
    Class representing a spectrum in memory.
    """
    def __init__(self):
        self.report: Optional[Report] = Report()
        self.id: str = new_spectrum_id_initialiser()
        self.database_id: int = NO_ID
        self.points: List[SpectrumPoint] = []
        self.max_amplitude: Optional[SpectrumPoint] = None
        self.min_amplitude: Optional[SpectrumPoint] = None
        self.max_wavenumber: Optional[SpectrumPoint] = None
        self.min_wavenumber: Optional[SpectrumPoint] = None

    def invalidate_min_max(self):
        """
        Resets the min/max values so they can be recalculated.
        """
        self.max_amplitude = None
        self.min_amplitude = None
        self.max_wavenumber = None
        self.min_wavenumber = None

    def validate_min_max(self):
        """
        Recalculates the min/max values.
        """
        first = True

        for point in self:
            if first:
                self.min_amplitude = point
                self.max_amplitude = point
                self.min_wavenumber = point
                self.max_wavenumber = point
                first = False
            else:
                if point.wave_number > self.max_wavenumber.wave_number:
                    self.max_wavenumber = point
                if point.wave_number < self.min_wavenumber.wave_number:
                    self.min_wavenumber = point
                if point.amplitude > self.max_amplitude.amplitude:
                    self.max_amplitude = point
                if point.amplitude < self.min_amplitude.amplitude:
                    self.min_amplitude = point

    def has_report(self):
        """
        Whether this spectrum has a report for meta-data.

        :return:    True if it has a report, False if not.
        """
        return self.report is not None

    def __iter__(self):
        # Treat ourself as an iterator over our points
        return self.points.__iter__()

    def __getattribute__(self, item):
        # Make sure min/max are valid
        if item in {'min_wavenumber', 'max_wavenumber', 'min_amplitude', 'max_amplitude'}:
            if super().__getattribute__(item) is None:
                self.validate_min_max()

        return super().__getattribute__(item)

    def __setattr__(self, key, value):
        # Don't handle specially until init is complete
        if key not in self.__dict__:
            super().__setattr__(key, value)
            return

        # Override report ids
        if key == 'report':
            if value is not None:
                value.database_id = self.database_id
                value.set_id(self.id.replace("'", ""))

        # Keep report id matching this id
        if key == 'id':
            if self.report is not None:
                self.report.id = value

        super().__setattr__(key, value)


def new_spectrum_id_initialiser() -> str:
    # TODO: Make return the curent date-time as per reference implementation. csterling
    return ''



"""
Package for working with JSON/JSONSchema in an object-oriented manner.
"""
from ._BasicSchemaValidator import BasicSchemaValidator
from ._deep_copy import deep_copy
from ._error import JSONError
from ._typing import RawJSONElement, RawJSONNumber, RawJSONArray, RawJSONObject, RawJSONPrimitive, \
    is_raw_json_primitive, is_raw_json_null, is_raw_json_number, is_raw_json_object, is_raw_json_array, \
    is_raw_json_element, is_raw_json_element_type



from ..._typing import RawJSONPrimitive
from ...schema import constant
from ._RawProperty import RawProperty


class ConstantProperty(RawProperty):
    """
    Configuration property which validates a constant primitive value.
    """
    def __init__(self,
                 name: str,
                 value: RawJSONPrimitive,
                 *,
                 optional: bool = False):
        super().__init__(
            name,
            constant(value),
            optional=optional
        )



"""
Module for static and dynamic typing of JSON schema.
"""
from typing import Union, Dict

import jsonschema

from ...meta import does_not_raise
from .._typing import RawJSONObject

# The type of a schema (boolean schema are trivial)
JSONSchema = Union[RawJSONObject, bool]

# The type of a definition
JSONDefinitions = Dict[str, JSONSchema]


def is_schema(schema):
    """
    Checks if the given object is a valid schema.

    :param schema:  The schema to check.
    :return:        True if it is a schema,
                    False if not.
    """
    # Get JSONSchema to check the schema against the latest standard
    return does_not_raise(jsonschema.validators.validator_for(schema).check_schema(schema))



"""
Provides utility functions for working with multiprocessing.Pool objects.
"""
from multiprocessing import Pool, Manager
from multiprocessing.managers import SyncManager
from multiprocessing.synchronize import Barrier
from typing import List, Any, Callable


def num_processes(pool: Pool) -> int:
    """
    Returns the number of sub-processes maintained by the given Pool.

    :param pool:    The pool.
    :return:        The number of sub-processes.
    """
    return pool._processes


def run_on_all(pool: Pool,
               func: Callable[[Any], Any],
               *args, **kwargs) -> List[Any]:
    """
    Runs a function on all processes managed by the given pool. Returns a list
    of results, one for each sub-process, with no guarantees as to order.

    :param pool:                The process pool.
    :param func:                The task to run.
    :param args:                The positional arguments to the task function.
    :param kwargs:              The keyword arguments to the task function.
    :return:                    The results from the workers.
    """
    # Get the number of processes to send the task too
    process_count: int = num_processes(pool)

    # Create a manager process to control a barrier for all workers
    barrier_manager: SyncManager = Manager()
    barrier: Barrier = barrier_manager.Barrier(process_count)

    # Create a copy of the task for each worker process with the barrier
    async_results = [pool.apply_async(func=run_with_barrier,
                                      args=(barrier, func, *args),
                                      kwds=kwargs)
                     for _ in range(process_count)]

    # Return the results when all are available
    return [async_result.get() for async_result in async_results]


def run_with_barrier(barrier: Barrier,
                     func: Callable[[Any], Any],
                     *args, **kwargs) -> Any:
    """
    Runs the given function with the provided arguments, once
    a barrier has been passed.

    :param barrier:     The barrier to pass before running.
    :param func:        The function to run.
    :param args:        The positional arguments to the function.
    :param kwargs:      The keyword arguments to the function.
    :return:            The result of running the function.
    """
    # Pass the barrier first
    barrier.wait()

    # Run the function
    return func(*args, **kwargs)



from typing import Iterable, Union, Sequence, TypeVar

SequenceType = TypeVar("SequenceType", bound=Sequence)


def flatten(sequence: SequenceType) -> SequenceType:
    """
    Flattens a sequence into a new sequence of the same type.
    Expects that the sequence-type's __init__ method takes an
    iterator (as in list and tuple).

    :param sequence:    The sequence to flatten.
    :return:            A new sequence of the same type, but flattened.
    """
    # Use flattening iterator to initialise a new sequence of the same type
    from ..iterate import flatten
    return type(sequence)(flatten(sequence))


def normalise_index(index: int, length: Union[int, Sequence]) -> int:
    """
    Normalises the given list index into the valid range for a
    list of the given length. This way negative indices index
    backward from the end of the list.

    :param index:   The index to normalise.
    :param length:  The sequence length to normalise to, or a sequence.
    :return:        The normalised index.
    """
    # Get the length if we're given a sequence
    if not isinstance(length, int):
        length = len(length)

    return index % length


def extract_by_index(sequence: SequenceType, indices: Iterable[int]) -> SequenceType:
    """
    Extracts elements of the given sequence. Returned type is the
    same as the given sequence. Expects that the sequence-type's
    __init__ method takes an iterator (as in list and tuple).

    :param sequence:    The sequence to get elements from.
    :param indices:     The indices of the elements to get.
    :return:            The selected elements.
    """
    # Use extracting iterator to initialise a new sequence of the same type
    from ..iterate import extract_by_index
    return type(sequence)(extract_by_index(sequence, indices))


def binary_search(sequence: Sequence[int], target: int) -> int:
    """
    Performs as binary search on an ordered sequence of ints to
    find the index of the first value greater than or equal to
    the target value.

    :param sequence:    The order sequence of ints.
    :param target:      The target value to find.
    :return:            The index of the found element.
    """
    start = 0
    end = len(sequence)
    while end != start:
        pivot = (start + end) // 2
        if sequence[pivot] >= target:
            end = pivot
        else:
            start = pivot + 1

    return start



"""
Package for the different types of properties a JSON configuration can have.
"""
from ._AnyOfProperty import AnyOfProperty
from ._ArrayProperty import ArrayProperty
from ._BoolProperty import BoolProperty
from ._ConstantProperty import ConstantProperty
from ._EnumProperty import EnumProperty
from ._MapProperty import MapProperty
from ._NumberProperty import NumberProperty
from ._OneOfProperty import OneOfProperty
from ._Property import Property
from ._ProxyProperty import ProxyProperty
from ._RawProperty import RawProperty
from ._StringProperty import StringProperty



"""
Typing functionality and static types for JSON package.
"""
from typing import Dict, Union, List, Any, Set

from ..iterate import all_meet_predicate

# The type of a raw JSON object in Python-land
RawJSONObject = Dict[str, 'RawJSONElement']

# The type of a raw JSON array in Python-land
RawJSONArray = List['RawJSONElement']

# The type of an arbitrary raw JSON number in Python-land
RawJSONNumber = Union[int, float]

# The type of a raw JSON primitive
RawJSONPrimitive = Union[RawJSONNumber, str, bool, None]

# The type of an arbitrary raw JSON element in Python-land
RawJSONElement = \
    Union[
        RawJSONObject,  # Object
        RawJSONArray,  # Array
        str,  # String
        RawJSONNumber,  # Number (real/integer)
        bool,  # True/false
        None  # Null
    ]

# The types of raw JSON elements
RAW_JSON_ELEMENT_TYPES: Set[type] = {dict, list, float, int, str, bool, None}


def is_raw_json_object(py_obj: Any) -> bool:
    """
    Checks if the given Python object is a raw JSON object.

    :param py_obj:  The Python object to check.
    :return:        True if the Python object is a raw JSON object,
                    False if not.
    """
    # Must be a dictionary
    if not isinstance(py_obj, dict):
        return False

    # All keys must be strings
    if not all(isinstance(key, str) for key in py_obj.keys()):
        return False

    # All values must be raw JSON elements
    if not all_meet_predicate(py_obj.values(), is_raw_json_element):
        return False

    # All checks passed
    return True


def is_raw_json_array(py_obj: Any) -> bool:
    """
    Checks if the given Python object is a raw JSON array.

    :param py_obj:  The Python object to check.
    :return:        True if the Python object is a raw JSON array,
                    False if not.
    """
    # Must be a list
    if not isinstance(py_obj, list):
        return False

    # All elements must be raw JSON elements
    if not all_meet_predicate(py_obj, is_raw_json_element):
        return False

    # All checks passed
    return True


def is_raw_json_number(py_obj: Any) -> bool:
    """
    Checks if the given Python object is a raw JSON number.

    :param py_obj:  The Python object to check.
    :return:        True if the Python object is a raw JSON number,
                    False if not.
    """
    # Must be an int or a float
    return isinstance(py_obj, int) or isinstance(py_obj, float)


def is_raw_json_null(py_obj: Any) -> bool:
    """
    Checks if the given Python object is a raw JSON null.

    :param py_obj:  The Python object to check.
    :return:        True if the Python object is a raw JSON null,
                    False if not.
    """
    # Must be None
    return py_obj is None


def is_raw_json_primitive(py_obj: Any) -> bool:
    """
    Checks if the given Python object is a raw JSON primitive.

    :param py_obj:  The Python object to check.
    :return:        True if the Python object is a raw JSON primitive,
                    False if not.
    """
    # Must be a number, string, bool or null
    return \
        is_raw_json_number(py_obj) or \
        isinstance(py_obj, str) or \
        isinstance(py_obj, bool) or \
        is_raw_json_null(py_obj)


def is_raw_json_element(py_obj: Any) -> bool:
    """
    Checks if the given Python object is any raw JSON element.

    :param py_obj:  The Python object to check.
    :return:        True if the Python object is a raw JSON element,
                    False if not.
    """
    return \
        is_raw_json_primitive(py_obj) or \
        is_raw_json_array(py_obj) or \
        is_raw_json_object(py_obj)


def is_raw_json_element_type(py_type: type) -> bool:
    """
    Checks if the given type is a raw JSON primitive type.

    :param py_type:     The type to check.
    :return:            True if the type is a raw JSON primitive type,
                        False if not.
    """
    return py_type in RAW_JSON_ELEMENT_TYPES



import io

from ._JSONSerialisable import JSONSerialisable
from ._JSONDeserialisable import JSONDeserialisable, T
from ._error import JSONSerialiseError


class JSONBiserialisable(JSONSerialisable, JSONDeserialisable[T]):
    """
    Interface for classes which implement both JSONSerialisable and
    JSONDeserialisable[T].
    """
    def json_copy(self) -> T:
        """
        Creates a copy of this object by serialising to JSON and
        then deserialising again.

        :return:    The copy of this object.
        """
        try:
            return self.from_raw_json(self.to_raw_json())
        except Exception as e:
            raise JSONSerialiseError("Error copying object using JSON serialisation") from e



from abc import abstractmethod
from typing import Tuple, Optional, List
from weakref import WeakKeyDictionary

from ...schema import JSONSchema
from ..._typing import RawJSONElement
from .._OptionallyPresent import OptionallyPresent
from .._Absent import Absent
from ._Property import Property
from ._DummyInstance import DummyInstance


class OfProperty(Property):
    """
    Base class for OneOfProperty and AnyOfProperty.
    """
    def __init__(self,
                 name: str,
                 *sub_properties: Property,
                 schema_function,  # one_of/any_of
                 optional: bool = False):
        super().__init__(
            name,
            optional=optional
        )

        self._sub_properties: Tuple[Property] = sub_properties
        self._instance_keys = WeakKeyDictionary()
        self.__schema_function = schema_function

    def get_json_validation_schema(self) -> JSONSchema:
        return self.__schema_function(
            *(
                property.get_json_validation_schema()
                for property in self._sub_properties
            )
        )

    def set_search_with_revert(self, instance, value, set_method):
        """
        Attempts to set this property, and reverts to the current
        value if setting fails.

        :param instance:    The instance to set the value for.
        :param value:       The value to set.
        :param set_method:  A function which returns the set-method
                            to use on an object.
        """
        # Make sure instance is not None
        self._require_instance(instance)

        # Attempt to get the current value
        try:
            current_value = self.__get__(instance, None)
        except AttributeError as e:
            # No current value
            current_value = e

        # Try to set the value
        try:
            self.set_search(instance, value, set_method)
        except Exception as e:
            if not isinstance(current_value, AttributeError):
                self.__set__(instance, current_value)

            raise e

    def set_search(self, instance, value, set_method):
        """
        Performs the logic of setting the value for this property.

        :param instance:    The instance to set the value for.
        :param value:       The value to set.
        :param set_method:  A function which returns the set-method
                            to use on an object.
        """
        # If the value is absent, set it directly
        if value is Absent:
            # Remove the current instance key if there is one
            if instance in self._instance_keys:
                self._instance_keys.pop(instance)

            super().__set__(instance, Absent)

            return

        # Find the properties this value is valid for
        keys = []
        for property in self._sub_properties:
            try:
                # Create a new key for this value
                keys.append(DummyInstance())

                set_method(property)(keys[-1], value)
            except Exception:
                keys[-1] = None

        # Select the canonical index
        key_index = self.choose_current_property(keys)

        # Save it
        super().__set__(instance, key_index)
        self._instance_keys[instance] = keys[key_index]

    @abstractmethod
    def choose_current_property(self, keys: List[Optional[DummyInstance]]) -> int:
        """
        Allows the sub-classes to choose which sub-property should
        store the value based on the success or failure of all sub-
        properties being set. Should raise an attribute error if it
        can't decide.

        :param keys:    The list of keys for the properties that were set,
                        and Nones for the properties that weren't.
        :return:        The index of the property to regard as storing the
                        value.
        """
        pass

    def __get__(self, instance, owner):
        # Get the current value
        value = super().__get__(instance, owner)

        # If it's an index, return the value of the sub-property
        # at that index
        if isinstance(value, int):
            return self._sub_properties[value].__get__(self._instance_keys[instance], owner)

        # Otherwise it's absent or self
        return value

    def __set__(self, instance, value):
        self.set_search_with_revert(instance,
                                    value,
                                    lambda property: property.__set__)

    def get_as_raw_json(self, instance) -> OptionallyPresent[RawJSONElement]:
        # Make sure instance is not None
        self._require_instance(instance)

        # Get the current value
        value = super().__get__(instance, None)

        # If it's an index, return the JSON of the sub-property
        # at that index
        if isinstance(value, int):
            return self._sub_properties[value].get_as_raw_json(self._instance_keys[instance])

        # Otherwise it's absent or self
        return value

    def set_from_raw_json(self, instance, value: OptionallyPresent[RawJSONElement]):
        self.set_search_with_revert(instance,
                                    value,
                                    lambda property: property.set_from_raw_json)

    def validate_value(self, value):
        super().validate_value(value)

        # No need to continue validation if value is absent
        if value is Absent:
            return

        # Our values are indices into the sub-properties
        if not isinstance(value, int) or not (0 <= value < len(self._sub_properties)):
            raise AttributeError(f"Error in internal index for Of property. "
                                 f"Index should be integer in [0:{len(self._sub_properties)}), "
                                 f"got {value}")



from random import Random, shuffle
from typing import Iterable, Iterator, Optional, Sequence, Tuple, List, Callable

from .._typing import Predicate, T


def is_iterable(obj) -> bool:
    """
    Determines if the given object is an iterable.

    :param obj:     The object to check.
    :return:        True if the object is iterable,
                    False if not.
    """
    return safe_iter(obj) is not None


def safe_iter(obj) -> Optional[Iterator]:
    """
    Returns an iterator for the object if it's iterable,
    or None if it's not.

    :param obj:     The object to get the iterator for.
    :return:        The iterator, or None if it's not iterable.
    """
    try:
        return iter(obj)
    except Exception:
        return None


def flatten(iterable: Iterable) -> Iterator:
    """
    Gets an iterator for the given iterable, which will
    flatten any iterable elements into the iterator output.

    :param iterable:    The iterable.
    :return:            A flattening iterator over the iterable.
    """
    # Process each element of the iterable in turn
    for element in iterable:
        # Try to get an iterator for the element
        element_iter = safe_iter(element)

        # If the element isn't iterable, yield the element itself
        if element_iter is None:
            yield element

        # Otherwise yield each sub-element of the flattened element
        else:
            for sub_element in flatten(element_iter):
                yield sub_element


def invert_indices(indices: Iterable[int], size: int) -> Iterator[int]:
    """
    Returns the indices that are not in the given indices, up to the given size.

    :param indices:     The indices to invert.
    :param size:        The exclusive maximum index.
    :return:            A inverted indices.
    """
    index_set = set(indices)

    for i in range(size):
        if i not in index_set:
            yield i


def extract_by_index(sequence: Sequence, indices: Iterable[int]) -> Iterator:
    """
    Extracts elements of the given sequence.

    :param sequence:    The sequence to get elements from.
    :param indices:     The indices of the elements to get.
    :return:            The selected elements.
    """
    for index in indices:
        yield sequence[index]


def first(iterable: Iterable[T], predicate: Predicate[T]) -> Tuple[int, Optional[T]]:
    """
    Finds the first value in an iterable to meet a given predicate.

    :param iterable:    The iterable to search.
    :param predicate:   The predicate to match.
    :return:            The enumerated position of and element found.
    """
    for iteration, value in enumerate(iterable):
        if predicate(value):
            return iteration, value

    # Predicate not satisfied
    return -1, None


def random(iterator: Iterator[T], rand: Random = None) -> Iterator[T]:
    """
    Returns the elements of the given iterator in a random order, as
    determined by the given source of randomness.

    :param iterator:    The iterator to randomise.
    :param rand:      The source of randomness.
    :return:            An iterator over the same elements in random order.
    """
    # Buffer the elements of the iterator
    buffer: List[T] = list(iterator)

    # Shuffle them
    if rand is not None:
        rand.shuffle(buffer)
    else:
        shuffle(buffer)

    return iter(buffer)


def all_meet_predicate(iterable: Iterable[T], predicate: Predicate[T]) -> bool:
    """
    Version of all which checks values against a predicate.

    :param iterable:    An iterable of values.
    :param predicate:   A predicate which checks those values.
    :return:            True if all values meet the predicate,
                        False if not.
    """
    return all(map(predicate, iterable))


def any_meets_predicate(iterable: Iterable[T], predicate: Predicate[T]) -> bool:
    """
    Version of any which checks values against a predicate.

    :param iterable:    An iterable of values.
    :param predicate:   A predicate which checks those values.
    :return:            True if any value meets the predicate,
                        False if not.
    """
    return any(map(predicate, iterable))



from typing import Any

from .._typing import T

META_DATA_KEY: str = "__metadata"


def with_metadata(obj: T, key: str, value: Any) -> T:
    """
    Adds meta-data to an object.

    :param obj:     The object to add meta-data to.
    :param key:     The key to store the meta-data under.
    :param value:   The meta-data value to store.
    :return:        obj.
    """
    # Create the meta-data map
    if not hasattr(obj, META_DATA_KEY):
        try:
            setattr(obj, META_DATA_KEY, {})
        except AttributeError as e:
            raise ValueError(f"Cannot set meta-data against objects of type {obj.__class__.__name__}") from e

    # Put this mapping in the map
    getattr(obj, META_DATA_KEY)[key] = value

    return obj


def get_metadata(obj: Any, key: str) -> Any:
    """
    Gets the meta-data from an object.

    :param obj:     The object to get meta-data from.
    :param key:     The meta-data key to extract from.
    :return:        The value of the meta-data.
    """
    return getattr(obj, META_DATA_KEY)[key]


def has_metadata(obj: Any, key: str) -> bool:
    """
    Checks if the given object has meta-data associated with
    the given key.

    :param obj:     The object to check.
    :param key:     The meta-data key.
    :return:        True if there is meta-data associated with the
                    given key, False if not.
    """
    return hasattr(obj, META_DATA_KEY) and key in getattr(obj, META_DATA_KEY)



from abc import abstractmethod, ABC
from typing import Iterable, Optional, Dict, Union, Mapping

from ..._typing import RawJSONElement
from ...serialise import JSONValidatedBiserialisable
from ...schema import JSONSchema, standard_object
from ._Property import Property
from ._DummyInstance import DummyInstance


class MapProxy(JSONValidatedBiserialisable['ArrayProxy'], ABC):
    """
    Class which acts like a map, but validates its elements using a property.
    """
    def __init__(self, initial_values: Optional[Union[Iterable, Mapping]] = None, **kwargs):
        # The references to the list elements in the sub-property
        self.__key_map: Dict[str, DummyInstance] = {}

        # Add any initial values
        self.update(initial_values, **kwargs)

    @classmethod
    @abstractmethod
    def sub_property(cls) -> Property:
        """
        Gets the property which is used to validate elements of the map.
        """
        pass

    def _serialise_to_raw_json(self) -> RawJSONElement:
        return {name: self.sub_property().get_as_raw_json(key) for name, key in self.__key_map.items()}

    @classmethod
    def _deserialise_from_raw_json(cls, raw_json: RawJSONElement) -> 'MapProxy':
        # Create the instance
        instance = cls()

        # Deserialise each value
        for name, value in raw_json.items():
            # Create a dummy key
            key = DummyInstance()

            # Add it to the list
            instance.__key_map[name] = key

            # Add the value to the property
            cls.sub_property().set_from_raw_json(key, value)

        return instance

    @classmethod
    def get_json_validation_schema(cls) -> JSONSchema:
        return standard_object(additional_properties=cls.sub_property().get_json_validation_schema())

    # ------------ #
    # DICT METHODS #
    # ------------ #

    def clear(self):
        self.__key_map.clear()

    def copy(self):
        # TODO
        raise NotImplementedError(MapProxy.copy.__qualname__)

    @staticmethod
    def fromkeys(seq, value):
        # TODO
        raise NotImplementedError(MapProxy.fromkeys.__qualname__)

    def get(self, k: str, default=None):
        if k not in self.__key_map:
            return default

        return self[k]

    def items(self):
        return ((key, self[key]) for key in self.keys())

    def keys(self):
        return self.__key_map.keys()

    def pop(self, k, d=None):
        if k not in self:
            if d is None:
                raise KeyError()
            else:
                return d

        value = self[k]
        self.__key_map.pop(k)
        return value

    def popitem(self):
        key, instance = self.__key_map.popitem()

        return key, self.sub_property().__get__(instance, None)

    def setdefault(self, key, default):
        if key not in self:
            self[key] = default

        return self[key]

    def update(self, E=None, **F):
        if E is not None:
            if hasattr(E, "keys") and callable(E.keys):
                for k in E:
                    self[k] = E[k]
            else:
                for k, v in E:
                    self[k] = v

        for k in F:
            self[k] = F[k]

    def values(self):
        return (self[k] for k in self.keys())

    def __contains__(self, key: str) -> bool:
        return key in self.__key_map

    def __delitem__(self, key: str):
        del(self.__key_map[key])

    def __eq__(self, *args, **kwargs):
        # TODO
        raise NotImplementedError(MapProxy.__eq__.__qualname__)

    def __getitem__(self, y):
        return self.sub_property().__get__(self.__key_map[y], None)

    def __ge__(self, *args, **kwargs):
        # TODO
        raise NotImplementedError(MapProxy.__ge__.__qualname__)

    def __gt__(self, *args, **kwargs):
        # TODO
        raise NotImplementedError(MapProxy.__gt__.__qualname__)

    def __iter__(self):
        return iter(self.__key_map)

    def __len__(self):
        return len(self.__key_map)

    def __le__(self, *args, **kwargs):
        # TODO
        raise NotImplementedError(MapProxy.__le__.__qualname__)

    def __lt__(self, *args, **kwargs):
        # TODO
        raise NotImplementedError(MapProxy.__lt__.__qualname__)

    def __ne__(self, *args, **kwargs):
        # TODO
        raise NotImplementedError(MapProxy.__ne__.__qualname__)

    def __repr__(self):
        return str(dict(self.items()))

    def __setitem__(self, key: str, value):
        # Create a dummy reference
        instance = DummyInstance()

        # Set the value on the sub-property
        self.sub_property().__set__(instance, value)

        # Put the reference in the key-map
        self.__key_map[key] = instance

    def __str__(self):
        return self.to_json_string()



from numbers import Real
from statistics import median
from typing import Iterable, Sequence


def lower_quartile(data: Iterable[Real]) -> Real:
    """
    Calculates the lower quartile of an iterable of real values.

    :param data:    The real values to calculate the lower quartile of.
    :return:        The lower quartile.
    """
    ordered = sorted(data)

    return lower_quartile_sorted(ordered)


def lower_quartile_sorted(data: Sequence[Real]) -> Real:
    """
    Calculates the lower quartile of a sorted sequence of real values.

    :param data:    The sorted real values to calculate the lower quartile of.
    :return:        The lower quartile.
    """
    return median(data[:len(data) // 2])


def upper_quartile(data: Iterable[Real]) -> Real:
    """
    Calculates the upper quartile of an iterable of real values.

    :param data:    The real values to calculate the upper quartile of.
    :return:        The upper quartile.
    """
    ordered = sorted(data)

    return upper_quartile_sorted(ordered)


def upper_quartile_sorted(data: Sequence[Real]) -> Real:
    """
    Calculates the upper quartile of a sorted sequence of real values.

    :param data:    The sorted real values to calculate the upper quartile of.
    :return:        The upper quartile.
    """
    return median(data[-len(data) // 2:])


def interquartile_range(data: Iterable[Real]) -> Real:
    """
    Calculates the inter-quartile range of an iterable of real values.

    :param data:    The real values to calculate the inter-quartile range from.
    :return:        The upper quartile.
    """
    ordered = sorted(data)

    return upper_quartile_sorted(ordered) - lower_quartile_sorted(ordered)



"""
Utility functions for working with iterables and iterators.
"""
from ._ConstantIterator import ConstantIterator
from ._functions import safe_iter, flatten, invert_indices, extract_by_index, is_iterable, first, random, \
    all_meet_predicate, any_meets_predicate
from ._metadata import with_metadata, zip_metadata, get_metadata, unzip_metadata



from abc import abstractmethod, ABC
from typing import Iterable, Optional, List, Callable, Any, Iterator

from ..._typing import RawJSONElement
from ...serialise import JSONValidatedBiserialisable, JSONSerialisable
from ...schema import JSONSchema, regular_array
from ._Property import Property
from ._DummyInstance import DummyInstance


class ArrayProxy(JSONValidatedBiserialisable['ArrayProxy'], ABC):
    """
    Class which acts like an array, but validates its elements using a property.
    """
    def __init__(self, initial_values: Optional[Iterable] = None):
        # The references to the list elements in the sub-property
        self.__key_list: List[DummyInstance] = []

        # Add any initial values
        if initial_values is not None:
            self.__iadd__(initial_values)

    @classmethod
    @abstractmethod
    def sub_property(cls) -> Property:
        """
        Gets the property which is used to validate elements of the array.
        """
        pass

    @classmethod
    @abstractmethod
    def min_elements(cls) -> int:
        """
        The minimum-allowed length of this array.
        """
        pass

    @classmethod
    @abstractmethod
    def max_elements(cls) -> Optional[int]:
        """
        The maximum-allowed length of this array.
        """
        pass

    @classmethod
    @abstractmethod
    def unique_elements(cls) -> bool:
        """
        Whether the elements of this array have to be distinct
        from one another.
        """
        pass

    def _serialise_to_raw_json(self) -> RawJSONElement:
        return [self.sub_property().get_as_raw_json(key) for key in self.__key_list]

    @classmethod
    def _deserialise_from_raw_json(cls, raw_json: RawJSONElement) -> 'ArrayProxy':
        # Create the instance
        instance = cls()

        # Create an initial key list
        instance.__key_list = [DummyInstance() for _ in range(len(raw_json))]

        # Deserialise and add the elements of the list
        for key, value in zip(instance.__key_list, raw_json):
            cls.sub_property().set_from_raw_json(key, value)

        return instance

    @classmethod
    def get_json_validation_schema(cls) -> JSONSchema:
        return regular_array(
            cls.sub_property().get_json_validation_schema(),
            cls.min_elements(),
            cls.max_elements(),
            cls.unique_elements()
        )

    # ------------ #
    # LIST METHODS #
    # ------------ #

    def append(self, value):
        # Make sure we're not already at max length
        if len(self.__key_list) == self.max_elements():
            raise ValueError(f"Tried to append to list already of maximum size ({self.max_elements()})")

        # Make sure the unique-elements constraint isn't violated
        if self.unique_elements() and value in self:
            raise ValueError(f"Attempted to add non-unique element")

        # Create a new key reference
        key = DummyInstance()

        # Add the value to the sub-property
        self.sub_property().__set__(key, value)

        # Append the key to the key-list
        self.__key_list.append(key)

    def clear(self):
        # Make sure clearing the list wouldn't violate the minimum size
        if self.min_elements() > 0:
            raise RuntimeError(f"Cannot clear array when minimum size "
                               f"({self.min_elements()}) is greater than zero")

        # Clear the key-list
        self.__key_list.clear()

    def copy(self):
        # TODO
        raise NotImplementedError(ArrayProxy.copy.__qualname__)

    def count(self, value) -> int:
        # TODO
        raise NotImplementedError(ArrayProxy.count.__qualname__)

    def extend(self, iterable: Iterable):
        for value in iterable:
            self.append(value)

    def index(self, value, start: int = 0, stop: int = -1) -> int:
        # TODO
        raise NotImplementedError(ArrayProxy.index.__qualname__)

    def insert(self, index: int, value):
        # Make sure we're not already at max length
        if len(self.__key_list) == self.max_elements():
            raise ValueError(f"Tried to insert into list already of maximum size ({self.max_elements()})")

        # Make sure the unique-elements constraint isn't violated
        if self.unique_elements() and value in self:
            raise ValueError(f"Attempted to insert non-unique element")

        # Create a new key reference
        key = DummyInstance()

        # Add the value to the sub-property
        self.sub_property().__set__(key, value)

        # Insert the key into the key-list
        self.__key_list.insert(index, key)

    def pop(self, index: int = -1):
        # Make sure we're not already at min length
        if len(self.__key_list) == self.min_elements():
            raise ValueError(f"Tried to pop from list already of minimum size ({self.min_elements()})")

        # Get the key
        key = self.__key_list.pop(index)

        # Return the value for the key
        return self.sub_property().__get__(key, None)

    def remove(self, value):
        self.pop(self.index(value))

    def reverse(self):
        self.__key_list.reverse()

    def sort(self, *,
             key: Optional[Callable[[Any], Any]] = lambda k: k,
             reverse: bool = False):
        self.__key_list.sort(key=lambda k: key(self.sub_property().__get__(k, None)), reverse=reverse)

    def __add__(self, x: List) -> List:
        return self[:] + x

    def __contains__(self, value) -> bool:
        return any(self.__values_equal(value, self.sub_property().__get__(key, None)) for key in self.__key_list)

    def __delitem__(self, *args, **kwargs):
        # TODO
        raise NotImplementedError(ArrayProxy.__delitem__.__qualname__)

    def __eq__(self, *args, **kwargs):
        # TODO
        raise NotImplementedError(ArrayProxy.__eq__.__qualname__)

    def __getitem__(self, y):
        if isinstance(y, slice):
            return [self[i] for i in range(*y.indices(len(self)))]
        else:
            return self.sub_property().__get__(self.__key_list.__getitem__(y), None)

    def __ge__(self, *args, **kwargs):
        # TODO
        raise NotImplementedError(ArrayProxy.__ge__.__qualname__)

    def __gt__(self, *args, **kwargs):
        # TODO
        raise NotImplementedError(ArrayProxy.__gt__.__qualname__)

    def __iadd__(self, x: Iterable):
        for value in x:
            self.append(value)

    def __imul__(self, *args, **kwargs):
        # TODO
        raise NotImplementedError(ArrayProxy.__imul__.__qualname__)

    def __iter__(self) -> Iterator:
        return (self.sub_property().__get__(key, None) for key in self.__key_list)

    def __len__(self) -> int:
        return len(self.__key_list)

    def __le__(self, *args, **kwargs):
        # TODO
        raise NotImplementedError(ArrayProxy.__le__.__qualname__)

    def __lt__(self, *args, **kwargs):
        # TODO
        raise NotImplementedError(ArrayProxy.__lt__.__qualname__)

    def __mul__(self, *args, **kwargs):
        # TODO
        raise NotImplementedError(ArrayProxy.__mul__.__qualname__)

    def __ne__(self, *args, **kwargs):
        # TODO
        raise NotImplementedError(ArrayProxy.__ne__.__qualname__)

    def __repr__(self):
        return str(element for element in self)

    def __reversed__(self):
        return (self.sub_property().__get__(key, None) for key in reversed(self.__key_list))

    def __rmul__(self, *args, **kwargs):
        # TODO
        raise NotImplementedError(ArrayProxy.__rmul__.__qualname__)

    def __setitem__(self, index: int, value):
        # Make sure the unique-elements constraint isn't violated
        if self.unique_elements() and value in self and not self.__values_equal(self[index], value):
            raise ValueError(f"Attempted to set element to non-unique element")

        # Get the key for the item
        key = self.__key_list[index]

        self.sub_property().__set__(key, value)

    def __str__(self):
        return self.to_json_string()

    @staticmethod
    def __values_equal(v1, v2) -> bool:
        """
        Checks if two values are equivalent in JSON.

        :param v1:      The first value to check.
        :param v2:      The second value to check.
        :return:        True if they are equivalent,
                        False if not.
        """
        # Do simple check first
        if v1 == v2:
            return True

        # Compare the JSON representations of the values
        if isinstance(v1, JSONSerialisable) and isinstance(v2, JSONSerialisable):
            return v1.to_raw_json() == v2.to_raw_json()

        return False



"""
Package specifying interfaces for objects/classes which can perform JSON validation.
"""
from ._JSONValidatorClass import JSONValidatorClass
from ._JSONValidatorInstance import JSONValidatorInstance



"""
Package for meta-programming constructs.
"""
from ._functions import has_been_overridden, get_class_from_function, is_class_function, unbind, does_not_raise
from ._metadata import with_metadata, get_metadata, has_metadata



"""
Schema for validating JSON which require no dynamic parameters.
"""
from .constants import *

# Trivial schemas for passing and failing
TRIVIALLY_SUCCEED_SCHEMA = True
TRIVIALLY_FAIL_SCHEMA = False

# The schema for validating only the value null
NULL_SCHEMA = {TYPE_KEYWORD: NULL_TYPE}

# The schema for validating boolean values
BOOL_SCHEMA = {TYPE_KEYWORD: BOOL_TYPE}

# The schema for validating an integer
INT_SCHEMA = {TYPE_KEYWORD: INTEGER_TYPE}

# The schema for validating a float
FLOAT_SCHEMA = {TYPE_KEYWORD: NUMBER_TYPE}

# The schema for validating a plain string
STRING_SCHEMA = {TYPE_KEYWORD: STRING_TYPE}

# Schema for validating a purely-positive integer
POSITIVE_INTEGER_SCHEMA = {TYPE_KEYWORD: INTEGER_TYPE, MINIMUM_KEYWORD: 1}

# Schema for validating a non-negative integer
NON_NEGATIVE_INTEGER_SCHEMA = {TYPE_KEYWORD: INTEGER_TYPE, MINIMUM_KEYWORD: 0}

# Schema for validating a purely-negative integer
NEGATIVE_INTEGER_SCHEMA = {TYPE_KEYWORD: INTEGER_TYPE, MAXIMUM_KEYWORD: -1}

# Schema for validating a non-positive integer
NON_POSITIVE_INTEGER_SCHEMA = {TYPE_KEYWORD: INTEGER_TYPE, MAXIMUM_KEYWORD: 0}



"""
Package for interfaces supporting JSON serialisation.
"""
from ._error import JSONSerialiseError
from ._JSONBiserialisable import JSONBiserialisable
from ._JSONDecodable import JSONDecodable
from ._JSONDeserialisable import JSONDeserialisable
from ._JSONEncodable import JSONEncodable
from ._JSONSerialisable import JSONSerialisable
from ._JSONValidatedBiserialisable import JSONValidatedBiserialisable



from typing import Type, TypeVar

import typing_inspect

from ._get_argument_to_typevar_base import check_args


def get_argument_to_typevar(cls: Type, generic_base_class: Type, typevar: TypeVar):
    """
    Gets the argument given to a type variable parameterising
    a generic base class in a particular sub-class of that base.

    :param cls:                 The sub-class specifying the argument.
    :param generic_base_class:  The generic base-class specifying the type variable.
    :param typevar:             The type variable to get the argument for.
    :return:                    The argument to the type variable.
    """
    # Check the arguments
    typevar_index: int = check_args(cls, generic_base_class, typevar)

    # Get the decendency path from derived to base class
    bases = [cls]
    while bases[-1] is not generic_base_class:
        # Keep track of if we found a base
        base_found = False

        # Try and find a generic base
        for base in typing_inspect.get_generic_bases(bases[-1]):
            if issubclass(base, generic_base_class):
                bases.append(base)
                base_found = True
                break

        # If we didn't find a generic base, find a non-generic base
        if not base_found:
            for base in bases[-1].__bases__:
                if issubclass(base, generic_base_class):
                    bases.append(base)
                    break

    # Search the dependency path for the type variable's final argument
    arg = None
    while len(bases) > 1:
        # Get the arguments to the generic base class
        args = typing_inspect.get_args(bases[-2])

        # If no arguments are given, the signature stays the same
        if len(args) == 0:
            bases = bases[:-2] + bases[-1:]
            continue

        # Get the argument to this typevar
        arg = args[typevar_index]

        # If it's another type variable, keep looking for the argument to this
        # type variable
        if typing_inspect.is_typevar(arg):
            typevar_index = typing_inspect.get_parameters(bases[-2]).index(arg)
            bases = bases[:-1]
            continue

        # Otherwise return the argument to the type variable
        return arg

    return arg



from abc import abstractmethod, ABC
from typing import Callable

from ..schema import JSONSchema
from .._typing import RawJSONElement
from ._JSONValidatorBase import JSONValidatorBase
from ._JSONValidatorInstance import JSONValidatorInstance


class JSONValidatorClass(JSONValidatorInstance, ABC):
    """
    Mixin class which adds validation support for JSON at the class level. Derives
    from JSONValidatorInstance as instances can still access the class-level validation.
    """
    @classmethod
    def with_validation(cls, func: Callable[[], RawJSONElement]) -> Callable[[], RawJSONElement]:
        return JSONValidatorBase.with_validation(cls, func)

    @classmethod
    def validate_raw_json(cls, raw_json: RawJSONElement):
        JSONValidatorBase.validate_raw_json(cls, raw_json)

    @classmethod
    def is_valid_raw_json(cls, raw_json: RawJSONElement) -> bool:
        return JSONValidatorBase.is_valid_raw_json(cls, raw_json)

    @classmethod
    def validate_json_string(cls, json_string: str):
        JSONValidatorBase.validate_json_string(cls, json_string)

    @classmethod
    def is_valid_json_string(cls, json_string: str) -> bool:
        return JSONValidatorBase.is_valid_json_string(cls, json_string)

    @classmethod
    @abstractmethod
    def get_json_validation_schema(cls) -> JSONSchema:
        pass

    @classmethod
    def clone_json_validation_schema(cls) -> JSONSchema:
        return JSONValidatorBase.clone_json_validation_schema(cls)

    @classmethod
    def perform_special_json_validation(cls, raw_json: RawJSONElement):
        pass



from numbers import Real
from typing import Optional

from ._InvalidStateError import InvalidStateError


class Interval:
    """
    Class which represents a numeric interval.
    E.g. [-1:1) or (4.333:12.78]

    Types of interval:
        - Total: contains all numbers.
        - Empty: contains no numbers.
        - Open-Upper: Contains all numbers above a certain value.
        - Open-Lower: Contains all numbers below a certain value.
        - Singular: Contains a single number.
        - Exclude-Value: Contains all numbers except a single value.
        - Outside: Contains all numbers that are NOT between two values.
        - Inside: The standard interval. Contains all numbers between two values.

    Manual manipulation of the members of this class may result in an
    InvalidStateError. Prefer to use the static constructor methods instead.
    """
    def __init__(self,
                 lower: Optional[Real],
                 upper: Optional[Real],
                 lower_inclusive: bool = True,
                 upper_inclusive: bool = False):
        self._lower: Optional[Real] = lower
        self._upper: Optional[Real] = upper
        self._lower_inclusive: bool = lower_inclusive
        self._upper_inclusive: bool = upper_inclusive

    def __contains__(self, value: Real) -> bool:
        if self.is_total():
            return True
        elif self.is_empty():
            return False
        elif self.is_open_upper():
            return self._above_lower(value)
        elif self.is_open_lower():
            return self._below_upper(value)
        elif self.is_singular():
            return value == self._lower
        elif self.is_exclude_value():
            return value != self._lower
        elif self.is_outside():
            return self._above_lower(value) or self._below_upper(value)
        elif self.is_inside():
            return self._above_lower(value) and self._below_upper(value)
        else:
            # Invalid state
            raise InvalidStateError("Misconfigured Interval")

    def __str__(self):
        # Format string for lower bound
        lower: str = "[" if self._lower_inclusive else "("
        if self._lower is not None:
            lower += str(self._lower)

        # Format string for upper bound
        upper: str = str(self._upper) if self._upper is not None else ""
        upper += "]" if self._upper_inclusive else ")"

        # Format the string parts into the whole
        if self.is_outside():
            return upper + lower
        elif self.is_singular():
            return lower + "]"
        elif self.is_exclude_value():
            return lower + ")"
        else:
            return lower + ":" + upper

    def is_total(self) -> bool:
        """
        Whether this interval is a total interval (contains
        all points).
        """
        return self._lower is None and self._upper is None and self._upper_inclusive and self._lower_inclusive

    def is_empty(self) -> bool:
        """
        Whether this interval is an empty interval (contains
        no points).
        """
        return self._lower is None and self._upper is None and not self._upper_inclusive and not self._lower_inclusive

    def is_open_lower(self) -> bool:
        """
        Whether this interval is an open-lower interval.
        """
        return self._upper is not None and self._lower is None and self._lower_inclusive

    def is_open_upper(self) -> bool:
        """
        Whether this interval is an open-upper interval.
        """
        return self._lower is not None and self._upper is None and self._upper_inclusive

    def is_half_open(self) -> bool:
        """
        Whether this interval is half-open.
        """
        return self.is_open_lower() or self.is_open_upper()

    def is_singular(self) -> bool:
        """
        Whether this interval is a singular interval (only
        includes one value).
        """
        return self._lower is not None and self._upper is not None and self._lower == self._upper \
            and self._upper_inclusive and self._lower_inclusive

    def is_exclude_value(self) -> bool:
        """
        Whether this interval is an exclude-value interval
        (includes all values except one).
        """
        return self._lower is not None and self._upper is not None and self._lower == self._upper \
            and not self._upper_inclusive and not self._lower_inclusive

    def is_outside(self) -> bool:
        """
        Whether this interval is an outside interval.
        """
        return self._lower is not None and self._upper is not None and self._lower > self._upper

    def is_inside(self) -> bool:
        """
        Whether this interval is a standard inside interval. Included for completeness.
        """
        return self._lower is not None and self._upper is not None and self._upper > self._lower

    def inverse(self) -> 'Interval':
        """
        Gets the inverse interval to this one (contains all points
        this one doesn't).
        """
        return Interval(self._upper, self._lower, not self._upper_inclusive, not self._lower_inclusive)

    def _above_lower(self, value: Real) -> bool:
        """
        Returns whether the given value is included by the
        lower bound. Assumes the lower bound exists.

        :param value:   The value to check.
        :return:        True if the value is contained by the
                        lower bound, False if not.
        """
        return value >= self._lower if self._lower_inclusive else value > self._lower

    def _below_upper(self, value: Real) -> bool:
        """
        Returns whether the given value is included by the
        upper bound. Assumes the upper bound exists.

        :param value:   The value to check.
        :return:        True if the value is contained by the
                        upper bound, False if not.
        """
        return value <= self._upper if self._upper_inclusive else value < self._upper

    @staticmethod
    def total() -> 'Interval':
        """
        Creates a total interval.
        """
        return Interval(None, None, True, True)

    @staticmethod
    def empty() -> 'Interval':
        """
        Creates an empty interval.
        """
        return Interval(None, None, False, False)

    @staticmethod
    def open_upper(lower: Real, lower_inclusive: bool = True) -> 'Interval':
        """
        Creates an open-upper interval.

        :param lower:               The lower-bound of the interval.
        :param lower_inclusive:     Whether the lower bound is inclusive.
        """
        return Interval(lower, None, lower_inclusive, True)

    @staticmethod
    def open_lower(upper: Real, upper_inclusive: bool = False) -> 'Interval':
        """
        Creates an open-lower interval.

        :param upper:               The upper bound of the interval.
        :param upper_inclusive:     Whether the upper bound is inclusive.
        """
        return Interval(None, upper, True, upper_inclusive)

    @staticmethod
    def singular(value: Real) -> 'Interval':
        """
        Creates a singular interval.

        :param value:   The single contained value.
        """
        return Interval(value, value, True, True)

    @staticmethod
    def exclude_value(value: Real) -> 'Interval':
        """
        Creates an exclude-value interval.

        :param value:   The single excluded value.
        """
        return Interval(value, value, False, False)

    @staticmethod
    def inside(value1: Real, value2: Real, lower_inclusive: bool = True, upper_inclusive: bool = False) -> 'Interval':
        """
        Creates an inside interval from possibly-unordered values.

        :param value1:              One bound of the inside interval.
        :param value2:              The other bound of the inside interval.
        :param lower_inclusive:     Whether the lower bound should be inclusive.
        :param upper_inclusive:     Whether the upper bound should be inclusive.
        """
        if value1 == value2:
            raise ValueError("Require unequal bounds for inside interval")

        return Interval(min(value1, value2), max(value1, value2), lower_inclusive, upper_inclusive)

    @staticmethod
    def outside(value1: Real, value2: Real, lower_inclusive: bool = True, upper_inclusive: bool = False) -> 'Interval':
        """
        Creates an outside interval from possibly-unordered values.

        :param value1:              One bound of the outside interval.
        :param value2:              The other bound of the outside interval.
        :param lower_inclusive:     Whether the lower bound should be inclusive.
        :param upper_inclusive:     Whether the upper bound should be inclusive.
        """
        if value1 == value2:
            raise ValueError("Require unequal bounds for outside interval")

        return Interval(max(value1, value2), min(value1, value2), lower_inclusive, upper_inclusive)



import json
from abc import ABC, abstractmethod
from typing import Callable

import jsonschema

from .._typing import RawJSONElement
from .._error import JSONError
from .._deep_copy import deep_copy
from ..schema import JSONSchema
from ...meta import does_not_raise


class JSONValidatorBase(ABC):
    """
    Provides base functionality for JSON validator interfaces.
    """
    @staticmethod
    def with_validation(self, func: Callable[[], RawJSONElement]) -> Callable[[], RawJSONElement]:
        """
        Decorator which adds validation to a function returning raw JSON.

        :param self:    The validator instance/class.
        :param func:    The function to decorate.
        :return:        The decorated function.
        """
        def validated() -> RawJSONElement:
            # Call the function
            raw_json: RawJSONElement = func()

            # Perform validation
            self.validate_raw_json(raw_json)

            return raw_json

        return validated

    @staticmethod
    def validate_raw_json(self, raw_json: RawJSONElement):
        """
        Validates the raw JSON using this object's validation methods
        (schema, special).

        :param self:        The validator instance/class.
        :param raw_json:    The raw JSON to validate.
        """
        try:
            # Perform schema validation
            jsonschema.validate(raw_json, self.get_json_validation_schema())

            # Perform special validation
            self.perform_special_json_validation(raw_json)
        except Exception as e:
            raise JSONError("Error validating raw JSON element") from e

    @staticmethod
    def is_valid_raw_json(self, raw_json: RawJSONElement) -> bool:
        """
        Checks if the raw JSON is valid.

        :param self:        The validator instance/class.
        :param raw_json:    The raw JSON to check.
        :return:            True if the JSON is valid,
                            False if not.
        """
        return does_not_raise(self.validate_raw_json, raw_json)

    @staticmethod
    def validate_json_string(self, json_string: str):
        """
        Validates the JSON-format string using this object's validation methods
        (schema, special).

        :param self:            The validator instance/class.
        :param json_string:     The JSON-format string to validate.
        """
        try:
            # Convert the string to raw JSON
            raw_json = json.loads(json_string)

            # Validate the raw JSON
            self.validate_raw_json(raw_json)
        except Exception as e:
            raise JSONError(f"Error validating JSON-format string {json_string}") from e

    @staticmethod
    def is_valid_json_string(self, json_string: str) -> bool:
        """
        Checks if the JSON-format string is valid.

        :param self:            The validator instance/class.
        :param json_string:     The JSON-format string to check.
        :return:                True if the JSON is valid,
                                False if not.
        """
        return does_not_raise(self.validate_json_string, json_string)

    @staticmethod
    @abstractmethod
    def get_json_validation_schema(self) -> JSONSchema:
        """
        Gets the schema to validate JSON for objects of this type.
        This may be a cached object, so if it is necessary to modify
        the schema, use clone_json_validation_schema instead.

        :param self:    The validator instance/class.
        :return:        The schema.
        """
        pass

    @staticmethod
    def clone_json_validation_schema(self) -> JSONSchema:
        """
        Gets a copy of the validation schema for this validator.
        Use this if you need to modify the schema.

        :param self:    The validator instance/class.
        :return:        The copy of the schema.
        """
        return deep_copy(self.get_json_validation_schema())

    @staticmethod
    def perform_special_json_validation(self, raw_json: RawJSONElement):
        """
        Optional method to perform custom validation of raw JSON elements
        for objects of this type. Should throw an exception if the raw JSON
        element doesn't meet the special validation criteria. By default
        does nothing.

        :param self:        The validator instance/class.
        :param raw_json:    The raw JSON to validate.
        """
        pass



from typing import Optional, List, Dict

from ._PythonVersionChecker import PythonVersionChecker
from ._Python36Checker import Python36Checker
from ._Python37Checker import Python37Checker


class SupportedVersionChecker(PythonVersionChecker):
    """
    Checks if the version of Python is supported by this library.
    """
    supported_versions: Dict[str, PythonVersionChecker] = {
        "Python 3.6": Python36Checker(),
        "Python 3.7": Python37Checker()
    }

    def check(self, major: int, minor: int, micro: int, releaselevel: str, serial: int) -> Optional[str]:
        if any(checker.check(major, minor, micro, releaselevel, serial) is None
               for checker in self.supported_versions.values()):
            return None
        return f"Version {major}.{minor}.{micro} {releaselevel} #{serial} is not supported. " \
            f"Supported versions: {', '.join(self.supported_versions.keys())}"



from ._Property import Property
from ._ProxyProperty import ProxyProperty
from ._MapProxy import MapProxy


class MapProperty(ProxyProperty):
    """
    Property which validates a map of strings to some other type.
    """
    def __init__(self,
                 sub_property: Property):

        # Create a closure class to proxy the array
        class ClosureMapProxy(MapProxy):
            @classmethod
            def sub_property(cls) -> Property:
                return sub_property

        super().__init__(
            sub_property.name(),
            ClosureMapProxy,
            optional=sub_property.is_optional()
        )

    def __set__(self, instance, value):
        # Convert other proxy-maps and raw dicts to our proxy-type
        if (isinstance(value, MapProxy) and not isinstance(value, self.type())) or \
           isinstance(value, dict):
            value = self.type()(value)

        super().__set__(instance, value)



from abc import ABC, abstractmethod
from typing import Optional, Union, Type, List, Tuple, Iterator

from ._save_helper import quoted
from ._error import ARFFError
from ..._TwoWayDict import TwoWayDict
from . import constants
from ._load_helper import remove_keyword, consume, is_whitespace_only, NAME_PATTERN, UnrecognisedContentError, \
    ATTRIBUTE_TYPE_PATTERN, DATE_FORMAT_PATTERN, NOMINAL_VALUE_PATTERN, remove_quotes


class Attribute(ABC):
    def __init__(self, name: str):
        self.name: str = name

    @abstractmethod
    def copy(self) -> 'Attribute':
        """
        Creates a copy of this attribute.
        """
        pass

    @classmethod
    @abstractmethod
    def type_string(cls) -> str:
        """
        Returns the type-string for this attribute type.
        """
        pass

    @classmethod
    def is_type(cls, *types: Union[str, Type['Attribute']]) -> bool:
        """
        Returns whether this attribute is one of the given types.

        :param types:   The types to check for, each as a string or class.
        :return:        True if the attribute is one of the given types,
                        False if not.
        """
        if len(types) == 0:
            return False
        else:
            for t in types:
                if isinstance(t, str) and cls.type_string() == t:
                    return True
                elif issubclass(cls, t):
                    return True

        return False

    def parse_string(self, string: str):
        """
        Parses a string into the correct internal type for this attribute.

        :param string:  The string to parse.
        :return:        The value for this attribute.
        """
        # Treat the missing value symbol as None
        if string == constants.MISSING_VALUE_SYMBOL:
            return None

        # Otherwise defer to the sub-type
        return self._parse_value(string)

    @abstractmethod
    def _parse_value(self, value):
        """
        Parses an object into the correct type for this attribute.

        :param value:    The value to parse.
        :return:         The internal type representation of the value.
        """
        pass

    @staticmethod
    def from_string(line: str) -> 'Attribute':
        """
        Parses an attribute definition string in ARFF format into an
        attribute object.

        :param line:    The string to parse.
        :return:        An attribute object.
        """
        return parse_line(line)

    def to_string(self) -> str:
        """
        Returns the ARFF format string specification of this attribute.
        """
        additional: str = self.to_string_additional()

        if additional != "":
            additional = " " + additional

        return constants.ATTRIBUTE_SECTION_KEYWORD + " " + quoted(self.name) + additional

    @abstractmethod
    def to_string_additional(self) -> str:
        """
        Returns any additional part of the string representation of this attribute
        after the attribute label and the attribute's name.
        """
        pass

    def __call__(self, value):
        # Missing value is valid for any attribute type
        if value is None:
            return None

        # Otherwise defer to the sub-type
        return self._parse_value(value)

    def __str__(self) -> str:
        return self.to_string()


class NumericAttribute(Attribute):
    """
    Class for numeric attributes.
    """
    def __init__(self, name: str, sub_type: Optional[str] = None):
        super().__init__(name)

        # Check the sub-type is valid
        if sub_type not in {None, constants.REAL_ATTRIBUTE_KEYWORD, constants.INTEGER_ATTRIBUTE_KEYWORD}:
            raise InvalidNumericSubTypeError(sub_type)

        self.sub_type: Optional[str] = sub_type

    def copy(self) -> 'NumericAttribute':
        return NumericAttribute(self.name, self.sub_type)

    @classmethod
    def type_string(cls) -> str:
        return constants.NUMERIC_ATTRIBUTE_KEYWORD

    def _parse_value(self, value) -> float:
        try:
            return float(value)
        except Exception as e:
            raise InvalidValueTypeError(value, self) from e

    def to_string_additional(self) -> str:
        return self.sub_type if self.sub_type is not None else self.type_string()


class NominalAttribute(Attribute):
    """
    Class for nominal attributes.
    """
    def __init__(self, name: str, values: List[str]):
        super().__init__(name)

        self.values: TwoWayDict[int, str] = TwoWayDict[int, str]()
        for index, value in enumerate(values):
            self.values[index] = value

    def copy(self) -> 'NominalAttribute':
        return NominalAttribute(self.name, self.ordered_values())

    def ordered_values(self) -> List[str]:
        """
        Gets a list of the nominal values in specification order.
        """
        return list(self.value_iterator())

    def value_iterator(self) -> Iterator[str]:
        """
        Gets an iterator over the nominal values.
        """
        return (self.values[index] for index in range(len(self.values)))

    def quoted_value_iterator(self) -> Iterator[str]:
        """
        Gets an iterator over the nominal values, quoting as necessary.
        """
        return (quoted(value) for value in self.value_iterator())

    @classmethod
    def type_string(cls) -> str:
        return constants.NOMINAL_ATTRIBUTE_KEYWORD

    def _parse_value(self, value) -> str:
        if isinstance(value, int):
            return self.values[value]
        elif isinstance(value, str):
            if value not in self.values:
                raise InvalidNominalValueError(value, self)

            return value
        else:
            raise InvalidValueTypeError(value, self)

    def to_string_additional(self) -> str:
        return "{" + ",".join(self.quoted_value_iterator()) + "}"

    def __iter__(self):
        return self.value_iterator()

    def __len__(self) -> int:
        return len(self.values)

    def __contains__(self, item: str) -> bool:
        return item in self.values


class StringAttribute(Attribute):
    """
    Class for string attributes.
    """
    def copy(self) -> 'StringAttribute':
        return StringAttribute(self.name)

    @classmethod
    def type_string(cls) -> str:
        return constants.STRING_ATTRIBUTE_KEYWORD

    def _parse_value(self, value) -> str:
        return str(value)

    def to_string_additional(self) -> str:
        return constants.STRING_ATTRIBUTE_KEYWORD


class DateAttribute(Attribute):
    """
    Class for date attributes.
    """
    def __init__(self, name: str, format_string: str):
        super().__init__(name)

        self.format: str = format_string

    def copy(self) -> 'DateAttribute':
        return DateAttribute(self.name, self.format)

    @classmethod
    def type_string(cls) -> str:
        return constants.DATE_ATTRIBUTE_KEYWORD

    def _parse_value(self, value) -> str:
        # TODO
        return value

    def to_string_additional(self) -> str:
        return constants.DATE_ATTRIBUTE_KEYWORD + " " + self.format


def parse_line(line: str) -> 'Attribute':
    """
    Parses an attribute line in ARFF file format into an attribute
    of the correct type.

    :param line:    The attribute line.
    :return:        The attribute.
    """
    # Save the original line for error messages
    original_line = line

    # Remove the @attribute keyword
    line = remove_keyword(line, constants.ATTRIBUTE_SECTION_KEYWORD)

    # Extract the attribute name
    name, line = consume(line, NAME_PATTERN)

    # Check we found a name
    if name is None:
        raise AttributeNameNotFoundError(original_line)

    # Extract the attribute type
    attribute_type, additional, line = consume_attribute_type_information(line)

    # Check we found a type
    if attribute_type is None:
        raise AttributeTypeNotFoundError(original_line)

    # If the line has not been entirely consumed, raise an error
    if not is_whitespace_only(line):
        raise UnrecognisedContentError(line, original_line)

    # Return the attribute and its type (with any additional info)
    return type_string_lookup[attribute_type](name, *additional)


def consume_attribute_type_information(line: str) -> \
        Tuple[Optional[str], Optional[Tuple[Union[str, List[str]]]], str]:
    """
    Extracts the type of an attribute from the beginning of a line.

    :param line:    The line to extract the attribute type from.
    :return:        The attribute type, any additional type information, and the remainder of the line.
    """
    # Find the attribute type
    attribute_type, line = consume(line, ATTRIBUTE_TYPE_PATTERN)

    # Abort if no match found
    if attribute_type is None:
        return None, None, line

    # Initialise no additional information
    additional = None

    # Add explicit nominal type
    if attribute_type.startswith('{'):
        additional = parse_nominal_values(attribute_type)
        attribute_type = constants.NOMINAL_ATTRIBUTE_KEYWORD

    # Lower-case the type
    attribute_type = attribute_type.lower()

    # Keep the date format string as additional
    if attribute_type == constants.DATE_ATTRIBUTE_KEYWORD:
        additional, line = consume(line, DATE_FORMAT_PATTERN)
    # Normalise numeric types
    elif attribute_type == constants.REAL_ATTRIBUTE_KEYWORD or \
            attribute_type == constants.INTEGER_ATTRIBUTE_KEYWORD:
        additional = attribute_type
        attribute_type = constants.NUMERIC_ATTRIBUTE_KEYWORD

    # Return the type information and the rest of the line
    return attribute_type, (additional,) if additional is not None else tuple(), line


def parse_nominal_values(string: str) -> List[str]:
    """
    Parses the string of nominal values into a Python list.

    :param string:  The string specification of nominal values (including braces).
    :return:        The list of nominal values.
    """

    # Save the original string for error messages
    original_string = string

    # Create the list
    values = []

    # Remove the leading brace
    string = string[1:]

    # Keep reading values until all are consumed
    while string != '}':
        # Remove any leading whitespace
        string = string.lstrip()

        # Get the name of the next value
        value, string = consume(string, NOMINAL_VALUE_PATTERN)

        # Make sure a value was found
        if value is None:
            raise NominalValuesError(original_string)

        # Put it in the list
        values.append(remove_quotes(value))

        # Remove any trailing whitespace after the value
        string = string.lstrip()

        # Remove the next delimiter
        if string[0] == ',':
            string = string[1:]
        elif string[0] != '}':
            raise NominalValuesError(original_string)

    # Return the list
    return values


# Create a lookup table between type-strings and Attribute classes
type_string_lookup: TwoWayDict[str, Type[Attribute]] = TwoWayDict[str, Type[Attribute]]()
for attribute_class in {NumericAttribute, NominalAttribute, StringAttribute, DateAttribute}:
    type_string_lookup[attribute_class] = attribute_class.type_string()


class InvalidNominalValueError(ARFFError):
    """
    Exception for when a nominal data value isn't one of the pre-specified
    values from the attribute declaration.
    """
    def __init__(self, value_string: str, attribute: NominalAttribute):
        super().__init__('Illegal nominal value for attribute \'' + attribute.name + '\': ' + value_string + '\n' +
                         'Allowed values are: ' + ', '.join(attribute.ordered_values()))


class AttributeNameNotFoundError(ARFFError):
    """
    Exception when a name for an attribute cannot be parsed.
    """
    def __init__(self, line: str):
        super().__init__("Couldn't parse attribute name from: " + line)


class AttributeTypeNotFoundError(ARFFError):
    """
    Exception for when an attribute's type cannot be determined.
    """
    def __init__(self, line: str):
        super().__init__("Couldn't determine attribute type from: " + line)


class NominalValuesError(ARFFError):
    """
    Exception for when the set of nominal values can't be parsed.
    """
    def __init__(self, string: str):
        super().__init__("Error parsing nominal values from: " + string)


class InvalidNumericSubTypeError(ARFFError):
    """
    Exception for using an invalid string as the sub-type of a numeric attribute.
    """
    def __init__(self, sub_type: str):
        super().__init__("'" + sub_type + "' is not a valid numeric sub-type.\n" +
                         "Allowed sub-types are: " + ", ".join((constants.REAL_ATTRIBUTE_KEYWORD,
                                                                constants.INTEGER_ATTRIBUTE_KEYWORD)))


class InvalidValueTypeError(ARFFError):
    """
    Exception for when a value can't be parsed into a value of the correct
    type for an attribute.
    """
    def __init__(self, value, attribute: Attribute):
        super().__init__("Attribute '" + attribute.name + "' of type '" + attribute.type_string() + "' can't " +
                         "convert value '" + str(value) + "' of type '" + str(value.__class__) + "'")



from typing import Optional, Tuple, Any, Dict, Type

from wai.test import AbstractTest
from wai.test.decorators import Test, ExceptionTest

from wai.common import TwoWayDict


class TwoWayDictTest(AbstractTest):
    @classmethod
    def subject_type(cls):
        return TwoWayDict[int, str]

    @Test
    def insert_and_retrieve(self, subject: TwoWayDict[int, str]):
        subject[2] = "asd"
        self.assertEqual(subject["asd"], 2)

    @ExceptionTest(TypeError)
    def insert_wrong_type(self, subject: TwoWayDict[int, str]):
        subject[3] = 3.4

    @Test
    def insert_class(self, subject: TwoWayDict[int, str]):
        class Base:
            pass

        class B1(Base):
            pass

        class B2(Base):
            pass

        new_subject = TwoWayDict[str, Type[Base]]()

        new_subject["B1"] = B1
        new_subject[B2] = "B2"

        self.assertIn(B1, new_subject)
        self.assertIn(B2, new_subject)
        self.assertIn("B1", new_subject)
        self.assertIn("B2", new_subject)
        self.assertEqual(new_subject[B1], "B1")
        self.assertEqual(new_subject["B2"], B2)



"""
Model for a pseudo-switch/case statement based around context-managers.

Use as follows:

with switch(value):
    if case(option1):
        ...
        break_()
    if case(option2):
        ...
        break_()
    if case(option3):
        ...
    if case(option4):
        ... falls through ...
        break_()
    if case(option5,
            option6,
            option7):
        ... multi-case ...
        break_()
    if default():
        ...
        break_()

"""
import threading
from typing import List, Dict

# The context-stack for each thread, so we can work out which
# switch statement the case function should work with
context_stacks: Dict[int, List['SwitchContextManager']] = dict()

# The thread lock for the context stacks
thread_lock: threading.RLock = threading.RLock()


def get_context_stack() -> List['SwitchContextManager']:
    """
    Makes sure the switch context-manager stack is initialised for this thread.

    :return:    The stack.
    """
    # Get the thread we're running on
    current_thread: int = threading.get_ident()

    with thread_lock:
        # Ensure there is a context stack for that thread
        if current_thread not in context_stacks:
            context_stacks[current_thread] = []

        # Return the stack
        return context_stacks[current_thread]


def add_to_stack(context_manager: 'SwitchContextManager'):
    """
    Adds a new SwitchContextManager to the thread stack.

    :param context_manager:     The new switch context manager.
    """
    # Put ourselves on the stack
    get_context_stack().append(context_manager)


def pop_from_stack():
    """
    Removes the top switch context manager from the thread stack.
    """
    # Get the context stack
    stack = get_context_stack()

    # Make sure it's not empty
    if len(stack) == 0:
        raise SwitchError("Attempted to remove switch context-manager from empty stack")

    # Pop the top of the stack
    stack.pop()


def get_current_context_manager() -> 'SwitchContextManager':
    """
    Gets the current switch context manager for this thread.

    :return:    The SwitchContextManager.
    """
    # Get the stack for this thread
    stack = get_context_stack()

    # Make sure it's not empty
    if len(stack) == 0:
        raise SwitchError("No switch context-manager for this thread")

    # Peek the top of the stack
    return stack[-1]


class SwitchContextManager:
    """
    Context manager for switch-case blocks.
    """
    def __init__(self, on):
        self.on = on
        self.matched: bool = False
        self.executed_default: bool = False

    def __enter__(self):
        # Add ourselves as the current context manager for the thread
        add_to_stack(self)

    def __exit__(self, exc_type, exc_val, exc_tb):
        # Remove ourselves as the current context manager for the thread
        pop_from_stack()

        # If a break statement was used to exit,
        # don't propagate the exception
        return exc_type is BreakSwitchException

    def case(self, first_case, *cases) -> bool:
        # Check we haven't already executed default
        self.check_default_executed()

        # If we match any of the provided cases, always match from now on
        if self.on in (first_case, *cases):
            self.matched = True

        # Allows for fall-through
        return self.matched

    def default(self) -> bool:
        # Check we haven't already executed default
        self.check_default_executed()

        # Flag that we've executed the default statement
        self.executed_default = True

        # Execute the default block if nothing has matched yet
        return not self.matched

    def check_default_executed(self):
        # If we've already executed default, automatically break
        if self.executed_default:
            break_()


def switch(on) -> SwitchContextManager:
    """
    Sugar method for setting up the switch-case context manager.

    :param on:  The value to be switched on.
    :return:    The switch context manager.
    """
    return SwitchContextManager(on)


def case(first_case, *cases) -> bool:
    """
    Sugar method which checks if any of the given cases matches
    the switch statement.

    :param first_case:  The first value to match.
    :param cases:       Any additional values to match.
    :return:            True if matched, False if not.
    """
    # Get the current context
    current_context: SwitchContextManager = get_current_context_manager()

    # Call it to check the cases provided
    return current_context.case(first_case, *cases)


def default():
    """
    Special version of the case statement which only matches
    if no previous case statements did. Automatically breaks.

    :return:    True if no case statement has matched,
                False otherwise.
    """
    # Get the current context
    current_context: SwitchContextManager = get_current_context_manager()

    # Check if we should execute the default block
    return current_context.default()


def break_():
    """
    Breaks from the switch block.
    """
    raise BreakSwitchException()


class BreakSwitchException(Exception):
    """
    A special exception that is thrown and caught by the switch context-manager
    to break from a switch block.
    """
    pass


class SwitchError(Exception):
    """
    Class for errors using the switch statement.
    """
    pass



class Encoding:
    """
    Class representing the one-hot encoding for a single set of strings.
    Typically represents one (source) column in a dataset.
    """
    def __init__(self, strings, numeric_type=int):
        # Set the numeric type
        self.__numeric_type = numeric_type

        # Create the lookup tables
        self.__lookup = {}
        self.__reverse_lookup = []

        # Add each unique string to the lookups
        for string in strings:
            if string not in self.__lookup:
                self.__lookup[string] = len(self.__lookup)
                self.__reverse_lookup.append(string)

        # Reset the internal encoding table
        self.__reset_encodings()

    def __reset_encodings(self):
        """
        Resets the internal encoding table cache to the empty state.
        Required when encoding state changes (e.g. numeric type).
        """
        self.__encodings = [None] * self.size()

    def set_numeric_type(self, numeric_type):
        """
        Sets the numeric type to use for encoded values.

        :param numeric_type:    The numeric type to use.
        """
        if numeric_type != self.__numeric_type and numeric_type is not None:
            self.__numeric_type = numeric_type
            self.__reset_encodings()

    def get_index(self, string):
        """
        Gets the encoded index for the given string.

        :param string:  The string to look up.
        :return:        The encoded index of the string.
        """

        # If the string is not encoded, return None
        if string not in self.__lookup:
            return None

        # Return the encoded index
        return self.__lookup[string]

    def get_string(self, index):
        """
        Returns the string that corresponds to the given encoded index.

        :param index:   The index to decode.
        :return:        The string that was encoded to the given index.
        """
        return self.__reverse_lookup[index]

    def get_encoding(self, string):
        """
        Gets the encoded array for the given string.

        :param string:  The string to encode.
        :return:        The encoded list, with a 1 at the encoded index and zeroes
                        everywhere else.
        """

        # Get the encoded index of the string
        index = self.get_index(string)

        # If the string wasn't encoded, abort
        if index is None:
            return None

        # If the encodings cache doesn't have an entry for this string, create it
        if self.__encodings[index] is None:
            self.__encodings[index] = create_encoded_array(index, self.size(), self.__numeric_type)

        # Return the encoding
        return self.__encodings[index]

    def encode_column(self, column):
        """
        Encodes an entire column of strings.

        :param column:  The list of strings to encode.
        :return:        A list of encodings.
        """
        return [self.get_encoding(string) for string in column]

    def get_header_names(self, prefix):
        """
        Gets a list of header names for the set of encoded columns produced
        by the given column name.

        :param prefix:  The name of the column being encoded.
        :return:        The list of column names produced by the encoding.
        """
        return [prefix + '_' + string for string in self.__reverse_lookup]

    def size(self):
        """
        Gets the number of encoded strings in this encoding.

        :return:    The number of encoded strings.
        """
        return len(self)

    def __len__(self):
        return len(self.__lookup)


class Mapping:
    """
    Class representing a collection of encodings, one for each column of a
    dataset. The entry for a particular column may be None, indicating that it
    should not be encoded.
    """
    def __init__(self, num_columns):
        # Create the empty collection of encodings
        self.__mapping = [None] * num_columns

    def add_encoding(self, column, encoding):
        """
        Adds an encoding to the collection at the specified index.

        :param column:      The column index of the encoded column.
        :param encoding:    The encoding.
        """
        self.__mapping[column] = encoding

    def encode_header(self, header):
        """
        Applies the encodings to the column names of a dataset.

        :param header:  The list of column names before encoding.
        :return:        The list of column names after encoding.
        """

        # Create the resultant list of column names
        encoded_header = []

        # Append each group of new column names
        for i, name in enumerate(header):
            if self.__mapping[i] is not None:
                for new_name in self.__mapping[i].get_header_names(name):
                    encoded_header.append(new_name)
            else:
                encoded_header.append(name)

        # Return the new column names
        return encoded_header

    def encode_column_selection(self, selection):
        """
        Applies the encodings to a list of selected column indices.

        :param selection:   The list of selected column indices.
        :return:            The encoded list of selected column indices.
        """

        # Create an accumulated list of encoded column lengths
        cumulative_lengths = [0]
        for i in range(0, len(self)):
            length = 1 if self.__mapping[i] is None else len(self.__mapping[i])
            cumulative_lengths.append(cumulative_lengths[i] + length)

        # Initialise the output
        output = []

        # For each selected index, add all of the indices between the accumulated
        # lengths (i.e. the range of encoded indices)
        for index in selection:
            for i in range(cumulative_lengths[index], cumulative_lengths[index + 1]):
                output.append(i)

        # Return the result
        return output

    def encode(self, data, row_major=True):
        """
        Encodes a dataset with the current encodings.

        :param data:        The dataset (a list of either rows or columns).
        :param row_major:   The data order.
        :return:            Nothing, the data is encoded in place.
        """
        if row_major:
            for row in data:
                self.encode_row(row)
        else:
            for i, column in enumerate(data):
                self.__mapping[i].encode_column(column)

    def encode_row(self, row):
        """
        Encodes a single row of data using the current set of encodings.

        :param row:     The row to encode.
        :return:        Nothing, the row is encoded in place.
        """

        # Process the row from last to first (to preserve indices after insertion)
        for i in reversed(range(len(row))):
            # Can skip non-encoded values
            if self.__mapping[i] is None:
                continue

            # Remove the string value
            value = row.pop(i)

            # Get the encoding from the map
            encoded_value = self.__mapping[i].get_encoding(value)

            # Insert the encoding into the data
            for j in reversed(range(len(encoded_value))):
                row.insert(i, encoded_value[j])

    def __len__(self):
        return len(self.__mapping)


def create_encoded_array(index, size, numeric_type=int):
    """
    Creates a one-hot encoded array with a one at the given index
    and zeroes everywhere else, up to the given size.

    :param index:           The index to set to one.
    :param size:            The size of the encoded array.
    :param numeric_type:    The numeric type for the values in the array.
    :return:        The one-hot encoding for the given parameters.
    """

    # Create the empty encoded value
    encoded_value = [numeric_type(0)] * size

    # Set the requested index to 1
    encoded_value[index] = numeric_type(1)

    # Return the encoded value
    return encoded_value



from typing import List, Optional

from ...schema import any_of
from ._OfProperty import OfProperty
from ._Property import Property
from ._DummyInstance import DummyInstance


class AnyOfProperty(OfProperty):
    """
    Property which validates JSON that matches at least one of
    a number of schema.
    """
    def __init__(self,
                 name: str,
                 *sub_properties: Property,
                 optional: bool = False):
        super().__init__(
            name,
            *sub_properties,
            schema_function=any_of,
            optional=optional
        )

    def choose_current_property(self, keys: List[Optional[DummyInstance]]) -> int:
        # Search the list
        for i, key in enumerate(keys):
            # Return the first valid index found
            if key is not None:
                return i

        raise AttributeError(f"Value didn't match any sub-properties")



from . import constants


def quoted(string: str) -> str:
    """
    Quotes a string if it contains spaces.

    :param string:  The string to optionally quote.
    :return:        The (possibly quoted) string.
    """
    return '"' + string + '"' if "'" in string else \
        "'" + string + "'" if '"' in string else \
        '"' + string + '"' if ' ' in string or \
                              string == constants.MISSING_VALUE_SYMBOL or \
                              constants.COMMENT_SYMBOL in string else \
        string



"""
Package for working with raw JSON schema.
"""
from ._definitions import IS_JSON_DEFINITION, IS_JSON_SCHEMA, IS_JSON_REFERENCE, IS_JSON_REFEREND, extract_definitions,\
    consolidate_definitions, reference_schema, reference_string
from ._dynamic import string_schema, standard_object, constant, enum, number, regular_array, one_of, any_of
from ._error import JSONSchemaError
from ._static import TRIVIALLY_SUCCEED_SCHEMA, TRIVIALLY_FAIL_SCHEMA, BOOL_SCHEMA, INT_SCHEMA, FLOAT_SCHEMA, \
    STRING_SCHEMA, POSITIVE_INTEGER_SCHEMA, NEGATIVE_INTEGER_SCHEMA, NON_NEGATIVE_INTEGER_SCHEMA, \
    NON_POSITIVE_INTEGER_SCHEMA
from ._typing import JSONSchema, JSONDefinitions, is_schema



from abc import ABC
from typing import Dict, TypeVar, List, Any, Union, Tuple, Type

from wai.common.json.configuration.property import RawProperty
from .property._MapProxy import MapProxy
from ._Absent import Absent
from ..schema import JSONSchema, standard_object, IS_JSON_SCHEMA, IS_JSON_DEFINITION
from ..schema.constants import DEFINITIONS_KEYWORD
from .property import Property
from .._typing import RawJSONObject
from ..serialise import JSONValidatedBiserialisable

T = TypeVar("T", bound="Configuration")


class Configuration(JSONValidatedBiserialisable[T], ABC):
    """
    Base class for Python representations of JSON-format configuration files.
    All configurations are validated against a schema for format correctness.
    The attributes of a configuration should be either JSON types or nested
    configurations.
    """
    # Cache of sub-class schema
    __schema_cache: Dict[Type["Configuration"], JSONSchema] = {}

    def __init__(self, **initial_values):
        # Get the properties
        properties = self.get_all_properties()

        # Make sure no properties are set twice
        for attribute_name, property in properties.items():
            if attribute_name in initial_values and \
                    property.name() in initial_values and \
                    property.name() not in properties:
                raise ValueError(f"Value for property {property.name()} (attribute "
                                 f"name '{attribute_name}') specified twice")

        # Set the properties with initial values, favouring attribute names
        for attribute_name, property in properties.items():
            # Get the initial value for the property (absent if not specified)
            initial_value = initial_values.pop(attribute_name
                                               if attribute_name in initial_values
                                               else property.name(),
                                               Absent)

            # Attempt to set the value
            try:
                # Set explicitly first
                property.__set__(self, initial_value)
            except Exception as e:
                try:
                    # If that fails, set by JSON
                    property.set_from_raw_json(self, initial_value)
                except Exception as e2:
                    # If that fails as well, let the exception escape
                    raise e2 from e

        # Treat remaining initial values as additional properties
        self._additional_properties = self.init_additional_properties(**initial_values)

    @classmethod
    def init_additional_properties(cls, **initial_values):
        """
        Initialises the additional-properties storage for this configuration.

        :param initial_values:  The initial values that weren't explicit properties.
        """
        # Create the closure of the sub-property
        sub_property: Property = cls._additional_properties_sub_property[0]

        # Create a map closure class to act as the storage
        class AdditionalPropertiesProxy(MapProxy):
            @classmethod
            def sub_property(cls) -> Property:
                return sub_property

        # Create the storage
        return AdditionalPropertiesProxy(**initial_values)

    def __getattr__(self, item: str):
        # Return the value of the additional property if present
        if item in self._additional_properties:
            return self._additional_properties[item]

        raise AttributeError(f"'{self.__class__.__name__}' has no attribute '{item}'")

    def __setattr__(self, key: str, value):
        # Set any unknown attribute in the additional values
        if key != "_additional_properties":
            if not hasattr(self, key) or key in self._additional_properties:
                self._additional_properties[key] = value

        super().__setattr__(key, value)

    def _serialise_to_raw_json(self) -> RawJSONObject:
        # Get a list of all properties
        properties = list(self.get_all_properties().values())

        # Convert the properties to JSON (includes Absent values)
        json_with_absents = {
            property.name(): property.get_as_raw_json(self)
            for property in properties
        }

        # Strip the absent values
        json = {
            key: value
            for key, value in json_with_absents.items()
            if value is not Absent
        }

        # Add any additional properties
        json.update(self._additional_properties)

        return json

    @classmethod
    def _deserialise_from_raw_json(cls, raw_json: RawJSONObject) -> T:
        return cls(**raw_json)

    @classmethod
    def get_json_validation_schema(cls) -> JSONSchema:
        # Get the cached version if possible
        if cls in Configuration.__schema_cache:
            return Configuration.__schema_cache[cls]

        # Get the properties of this configuration type
        properties: Dict[str, Property] = cls.get_all_properties(True)

        # Extract the required property schemas
        required_properties_schema: Dict[str, JSONSchema] = {
            name: property.get_json_validation_schema()
            for name, property in properties.items()
            if not property.is_optional()
        }

        # Extract the optional property schemas
        optional_properties_schema: Dict[str, JSONSchema] = {
            name: property.get_json_validation_schema()
            for name, property in properties.items()
            if property.is_optional()
        }

        # Extract the additional properties schema
        additional_properties_schema: JSONSchema = cls._additional_properties_sub_property[0]\
            .get_json_validation_schema()

        # Create the schema
        schema: JSONSchema = standard_object(required_properties_schema,
                                             optional_properties_schema,
                                             additional_properties=additional_properties_schema)

        # Cache the schema
        Configuration.__schema_cache[cls] = schema

        return schema

    @classmethod
    def get_all_properties(cls, with_property_names: bool = False) -> Dict[str, Property]:
        """
        Gets all properties of this configuration.

        :param with_property_names:     Whether the keys should be the property names (by default
                                        they are the attribute names).
        :return:                        A dictionary of the properties of this configuration.
        """
        # Get the attribute names
        attribute_names: List[str] = dir(cls)

        # Get the mapping from attribute name to attribute
        attributes: Dict[str, Any] = {name: getattr(cls, name)
                                      for name in attribute_names
                                      if not name.startswith("_")}

        # Get the properties
        properties = {
            name: attribute
            for name, attribute in attributes.items()
            if isinstance(attribute, Property)
        }

        # Convert the names to the property names if requested
        if with_property_names:
            properties = {property.name(): property for property in properties.values()}

        return properties

    @classmethod
    def additional_properties_validation(cls) -> Union[JSONSchema, Property]:
        """
        Gets the schema/property to validate any additional properties set
        on this configuration. By default all additional properties are allowed.
        Override to modify this behaviour.
        """
        schema: JSONSchema = {DEFINITIONS_KEYWORD: IS_JSON_DEFINITION}
        schema.update(IS_JSON_SCHEMA)
        return schema

    def __init_subclass__(cls, **kwargs):
        # Perform any super initialisation
        super().__init_subclass__(**kwargs)

        # Make sure the property names are unique
        property_names = [property.name() for property in cls.get_all_properties().values()]
        if len(property_names) != len(set(property_names)):
            raise TypeError(f"Duplicate property names in {cls.__name__}: {', '.join(property_names)}")

        # Initialise the additional properties validation property
        property = cls.additional_properties_validation()
        if not isinstance(property, Property):
            property = RawProperty("", property)
        cls._additional_properties_sub_property: Tuple[Property] = (property,)  # Wrapped in a tuple to avoid discovery



from ..._typing import RawJSONPrimitive
from ...schema import enum
from ._RawProperty import RawProperty


class EnumProperty(RawProperty):
    """
    Property which accepts one of a set of specific primitive values.
    """
    def __init__(self,
                 name: str,
                 *values: RawJSONPrimitive,
                 optional: bool = False):
        super().__init__(
            name,
            enum(*values),
            optional=optional
        )



"""
Methods for creating some common JSON-validation schema,
based on some parameters.
"""
from typing import Dict, Set, Optional, Union

from ._error import JSONSchemaError
from ._typing import JSONSchema
from .._typing import RawJSONNumber, RawJSONPrimitive
from .constants import *
from ._definitions import consolidate_definitions, JSONDefinitions
from .._deep_copy import deep_copy


def string_schema(min_length: Optional[int] = None,
                  max_length: Optional[int] = None,
                  pattern: Optional[str] = None,
                  format: Optional[str] = None) -> JSONSchema:
    """
    Creates a schema which validates a string.

    TODO: Format strings are only optionally enforced by validators. Add check
          to ensure that the given format string will be enforced, and possibly
          a flag to skip this check if unnecessary.

    :param min_length:  The minimum length of the string.
    :param max_length:  The maximum length of the string.
    :param pattern:     A regex to validate against.
    :param format:      The format type to validate against.
    :return:            The schema.
    """
    # Create the base schema
    schema: JSONSchema = {
        TYPE_KEYWORD: STRING_TYPE
    }

    # Add the minimum length if given
    if min_length is not None:
        if min_length < 0:
            raise ValueError(f"min_length must be non-negative if given, got {min_length}")
        schema[MIN_LENGTH_KEYWORD] = min_length

    # Add the maximum length if given
    if max_length is not None:
        if max_length < 0:
            raise ValueError(f"max_length must be non-negative if given, got {max_length}")
        schema[MAX_LENGTH_KEYWORD] = max_length

    # Add the pattern if given
    if pattern is not None:
        schema[PATTERN_KEYWORD] = pattern

    # Add the format if given
    if format is not None:
        schema[FORMAT_KEYWORD] = format

    return schema


def standard_object(required_properties: Optional[Dict[str, JSONSchema]] = None,
                    optional_properties: Optional[Dict[str, JSONSchema]] = None,
                    additional_properties: Union[JSONSchema, bool] = False) -> JSONSchema:
    """
    Creates an schema which validates an object.

    :param required_properties:             The required properties.
    :param optional_properties:             The optional properties.
    :param additional_properties:           Either a schema for validating any additional properties,
                                            or a boolean specifying whether additional properties are allowed.
    :return:                                The schema.
    """
    # Set the default properties if none are provided
    if required_properties is None:
        required_properties = {}
    if optional_properties is None:
        optional_properties = {}

    # Make sure there are not duplicate properties between required and optional.
    for key in required_properties:
        if key in optional_properties:
            raise JSONSchemaError(f"Property '{key}' can't be both required and optional")

    # Copy all schema
    required_properties = deep_copy(required_properties)
    optional_properties = deep_copy(optional_properties)
    additional_properties = deep_copy(additional_properties)

    # Extract any definitions
    definitions: JSONDefinitions = consolidate_definitions(*required_properties.values(),
                                                           *optional_properties.values(),
                                                           additional_properties,
                                                           pop=True)

    # Combine the required and optional properties
    combined_properties = {}
    combined_properties.update(required_properties)
    combined_properties.update(optional_properties)

    # Create the schema
    schema = {
        TYPE_KEYWORD: OBJECT_TYPE,
        PROPERTIES_KEYWORD: combined_properties,
        DEFINITIONS_KEYWORD: definitions
    }

    # Add the list of required properties if there are any
    if len(required_properties) > 0:
        schema[REQUIRED_PROPERTIES_KEYWORD] = list(required_properties.keys())

    # Disallow additional properties if not selected
    if additional_properties is not True:
        schema[ADDITIONAL_PROPERTIES_KEYWORD] = additional_properties

    return schema


def constant(value: RawJSONPrimitive) -> JSONSchema:
    """
    Creates a schema that only validates a constant value.

    :param value:   The constant value.
    :return:        The schema.
    """
    return {CONSTANT_KEYWORD: value}


def enum(*values: RawJSONPrimitive) -> JSONSchema:
    """
    Creates a schema that only validates a fixed set of values.

    :param values:  The set of allowed values.
    :return:        The schema.
    """
    unique_values: Set[RawJSONPrimitive] = set(values)

    return {ENUMERATION_KEYWORD: list(unique_values)}


def number(minimum: Optional[RawJSONNumber] = None,
           maximum: Optional[RawJSONNumber] = None,
           integer_only: bool = False,
           multiple_of: Optional[RawJSONNumber] = None,
           exclusive_minimum: bool = False,
           exclusive_maximum: bool = False) -> JSONSchema:
    """
    Creates a schema that validates a numeric value.

    :param minimum:             The minimum value the number must take.
    :param maximum:             The maximum value the number must take.
    :param integer_only:        Whether to only accept integer values.
    :param multiple_of:         The number the value must be a multiple of.
    :param exclusive_minimum:   Whether the minimum value is exclusive.
    :param exclusive_maximum:   Whether the maximum value is exclusive.
    :return:                    The schema.
    """
    # Create the schema with the correct type
    schema: JSONSchema = {
        TYPE_KEYWORD: NUMBER_TYPE if not integer_only else INTEGER_TYPE
    }

    # Set the minimum if supplied
    if minimum is not None:
        schema[MINIMUM_KEYWORD if not exclusive_minimum else EXCLUSIVE_MINIMUM_KEYWORD] = minimum

    # Set the maximum if supplied
    if maximum is not None:
        schema[MAXIMUM_KEYWORD if not exclusive_maximum else EXCLUSIVE_MAXIMUM_KEYWORD] = maximum

    # Set the multiple if supplied
    if multiple_of is not None:
        schema[MULTIPLE_OF_KEYWORD] = multiple_of

    return schema


def regular_array(element_schema: JSONSchema,
                  min_elements: int = 0,
                  max_elements: Optional[int] = None,
                  unique_elements: bool = False) -> JSONSchema:
    """
    Creates a schema validating an array of items of the same type.

    :param element_schema:  The schema to validate each item.
    :param min_elements:    The minimum number of elements that must be in the array.
    :param max_elements:    The maximum number of elements that must be in the array.
    :param unique_elements: Whether the elements of the array must be distinct.
    :return:                The schema.
    """
    # It's invalid to provided a negative value for an element count restriction
    if min_elements < 0 or (max_elements is not None and max_elements < 0):
        raise JSONSchemaError("Can't have an array with a negative number of elements")

    # Can't have a maximum less than a minimum
    if max_elements is not None and max_elements < min_elements:
        raise JSONSchemaError("max_elements can't be less than min_elements")

    # Copy the element schema
    element_schema = deep_copy(element_schema)

    # Extract any definitions
    definitions: JSONDefinitions = consolidate_definitions(element_schema, pop=True)

    # Create the schema with the array type and element schema
    schema: JSONSchema = {
        TYPE_KEYWORD: ARRAY_TYPE,
        ITEMS_KEYWORD: element_schema,
        DEFINITIONS_KEYWORD: definitions
    }

    # Add the minimum length if more than nothing
    if min_elements > 0:
        schema[MIN_ITEMS_KEYWORD] = min_elements

    # Add the maximum length if present
    if max_elements is not None:
        schema[MAX_ITEMS_KEYWORD] = max_elements

    # Add the uniqueness flag if set
    if unique_elements:
        schema[UNIQUE_ITEMS_KEYWORD] = True

    return schema


def one_of(*schema: JSONSchema) -> JSONSchema:
    """
    Creates a schema which makes sure the value matches only one of the
    given schema.

    :param schema:  The individual schema.
    :return:        The combined schema.
    """
    # Must provide at least 2 schema
    if len(schema) < 2:
        raise JSONSchemaError("Can't use one_of with fewer than 2 sub-schema")

    # Copy the schema into a list
    schema = deep_copy(list(schema))

    # Extract any definitions
    definitions: JSONDefinitions = consolidate_definitions(*schema, pop=True)

    return {
        ONE_OF_KEYWORD: schema,
        DEFINITIONS_KEYWORD: definitions
    }


def any_of(*schema: JSONSchema) -> JSONSchema:
    """
    Creates a schema which makes sure the value matches at least one of the
    given schema.

    :param schema:  The individual schema.
    :return:        The combined schema.
    """
    # Must provide at least 2 schema
    if len(schema) < 2:
        raise JSONSchemaError("Can't use one_of with fewer than 2 sub-schema")

    # Copy the schema into a list
    schema = deep_copy(list(schema))

    # Extract any definitions
    definitions: JSONDefinitions = consolidate_definitions(*schema, pop=True)

    return {
        ANY_OF_KEYWORD: schema,
        DEFINITIONS_KEYWORD: definitions
    }



"""
Module for constants related to JSON validation schema.
"""
# Keywords
ADDITIONAL_PROPERTIES_KEYWORD: str = "additionalProperties"  # The schema for validating any additional properties
ANY_OF_KEYWORD: str = "anyOf"  # For matching against any one of a number of schema
CONSTANT_KEYWORD: str = "const"  # For validating against a constant value
DEFINITIONS_KEYWORD: str = "definitions"  # For defining reference schema
ENUMERATION_KEYWORD: str = "enum"  # For validating against a set of constant values
EXCLUSIVE_MAXIMUM_KEYWORD: str = "exclusiveMaximum"  # The exclusive maximum of a numeric type
EXCLUSIVE_MINIMUM_KEYWORD: str = "exclusiveMinimum"  # The exclusive minimum of a numeric type
FORMAT_KEYWORD: str = "format"  # The custom format to match strings against
ITEMS_KEYWORD: str = "items"  # The schema to validate elements of an array
MAXIMUM_KEYWORD: str = "maximum"  # The maximum value of a numeric type
MAX_ITEMS_KEYWORD: str = "maxItems"  # The maximum number of items in an array
MAX_LENGTH_KEYWORD: str = "maxLength"  # The maximum length of a string
MINIMUM_KEYWORD: str = "minimum"  # The minimum value of a numeric type
MIN_ITEMS_KEYWORD: str = "minItems"  # The minimum number of items in an array
MIN_LENGTH_KEYWORD: str = "minLength"  # The minimum length of a string
MULTIPLE_OF_KEYWORD: str = "multipleOf"  # The divisor of numeric types
ONE_OF_KEYWORD: str = "oneOf"  # For validating against exactly one of a number of schema
PATTERN_KEYWORD: str = "pattern"  # The regex pattern to match against strings
PROPERTIES_KEYWORD: str = "properties"  # The schema for validating object properties
REFERENCE_KEYWORD: str = "$ref"  # For referring to reference schema
REQUIRED_PROPERTIES_KEYWORD: str = "required"  # The names of required object properties
TYPE_KEYWORD: str = "type"  # The type of JSON element to validate
UNIQUE_ITEMS_KEYWORD: str = "uniqueItems"  # Whether all elements in an array must be unique

# Types
ARRAY_TYPE: str = "array"
INTEGER_TYPE: str = "integer"
NUMBER_TYPE: str = "number"
OBJECT_TYPE: str = "object"
STRING_TYPE: str = "string"
BOOL_TYPE: str = "boolean"
NULL_TYPE: str = "null"



"""
Module for reusable JSON schema definitions and references to them.
Also provides tools for relative JSON pointer syntax.
"""
from .constants import *
from ._static import NULL_SCHEMA, BOOL_SCHEMA, FLOAT_SCHEMA, STRING_SCHEMA
from ._typing import JSONSchema, JSONDefinitions


def reference_string(referend: str) -> str:
    """
    Creates a JSON pointer reference string to the given referend.

    :param referend:    The referend.
    :return:            The reference string.
    """
    return f"#/{DEFINITIONS_KEYWORD}/{referend}"


def reference_schema(referend: str) -> JSONSchema:
    """
    Returns a reference schema to the given referend.

    :param referend:    The referend.
    :return:            The schema.
    """
    return {REFERENCE_KEYWORD: reference_string(referend)}


def extract_definitions(schema: JSONSchema, pop: bool = False) -> JSONDefinitions:
    """
    Extracts the definitions for the given JSON schema.

    :param schema:  The schema.
    :param pop:     Whether to pop the definitions or leave them in place.
    :return:        The definitions.
    """
    # Trivial schema have no definitions
    if isinstance(schema, bool):
        return {}

    # If there are no definitions, return an empty set
    if DEFINITIONS_KEYWORD not in schema:
        return {}

    # Get or pop the definitions as requested
    if pop:
        return schema.pop(DEFINITIONS_KEYWORD)
    else:
        return schema.get(DEFINITIONS_KEYWORD)


def consolidate_definitions(*schemata: JSONSchema, pop: bool = False) -> JSONDefinitions:
    """
    Consolidates the definitions from a number of schemata,
    checking for collisions.

    :param schemata:        The schemas to collect definitions from.
    :param pop:             Whether to pop the definitions or leave them in place.
    :return:                The definitions.
    """
    # Create the results set
    consolidated_definitions = {}

    # Process each schema in turn
    for schema in schemata:
        extracted_definitions = extract_definitions(schema, pop)
        for definition_name, definition in extracted_definitions.items():
            if definition_name in consolidated_definitions and consolidated_definitions[definition_name] != definition:
                raise ValueError(f"Error composing definitions: definition '{definition_name}' "
                                 f"defined more than once")
            consolidated_definitions[definition_name] = definition

    return consolidated_definitions


# Definition of a schema which validates any JSON
IS_JSON_REFERENCE: str = "is-json"
IS_JSON_REFEREND: str = reference_string(IS_JSON_REFERENCE)
IS_JSON_SCHEMA: JSONSchema = reference_schema(IS_JSON_REFERENCE)
IS_JSON_DEFINITION: JSONDefinitions = {
    IS_JSON_REFERENCE: {
        ANY_OF_KEYWORD: [
            NULL_SCHEMA,
            BOOL_SCHEMA,
            FLOAT_SCHEMA,
            STRING_SCHEMA,
            {TYPE_KEYWORD: OBJECT_TYPE, ADDITIONAL_PROPERTIES_KEYWORD: IS_JSON_SCHEMA},
            {TYPE_KEYWORD: ARRAY_TYPE, ITEMS_KEYWORD: IS_JSON_SCHEMA}
        ]
    }
}


from abc import ABC
from typing import TypeVar

from ..schema import JSONSchema, TRIVIALLY_FAIL_SCHEMA
from ._Configuration import Configuration

T = TypeVar("T", bound="StrictConfiguration")


class StrictConfiguration(Configuration[T], ABC):
    """
    Base class for configurations which don't support additional
    properties by default.
    """
    @classmethod
    def additional_properties_validation(cls) -> JSONSchema:
        return TRIVIALLY_FAIL_SCHEMA



from abc import ABC

from ..validator import JSONValidatorClass
from ._JSONBiserialisable import JSONBiserialisable, T


class JSONValidatedBiserialisable(JSONValidatorClass, JSONBiserialisable[T], ABC):
    """
    Utility interface that specifies that a class is both biserialisable
    with JSON and that that JSON is validated.
    """
    pass



"""
Defines the OptionallyPresent type, which can take values of
sub-script type T or the singleton Absent value.

E.g. a: OptionallyPresent[int] = |any int or Absent|
"""
from typing import TypeVar, Union

from ._Absent import AbsentType

# The type of the value if it is present
T = TypeVar('T')

OptionallyPresent = Union[AbsentType, T]
